from collections.abc import Sequence
from typing import Any, ClassVar, Literal, TypeVar
from typing_extensions import Self, override

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, Float, PyTree

from ..._internal import error_if_not_positive
from ...constants import LobatoScatteringFactorParameters, PengScatteringFactorParameters
from ...jax_util import FloatLike, NDArrayLike
from ...ndimage import (
    AbstractFourierOperator,
    FourierSinc,
    block_reduce_downsample,
    convert_fftn_to_rfftn,
    ifftn,
    irfftn,
    make_frequency_grid,
    resize_with_crop_or_pad,
    rfftn,
)
from .._image_config import AbstractImageConfig
from .._pose import AbstractPose
from .base_volume import (
    AbstractAtomVolume,
    AbstractVolumeIntegrator,
    AbstractVolumeRenderFn,
    ProjectionArray,
    VoxelArray,
)


try:
    import jax_finufft as jnufft

    JAX_FINUFFT_IMPORT_ERROR = None
except ModuleNotFoundError as err:
    jnufft = None
    JAX_FINUFFT_IMPORT_ERROR = err


T = TypeVar("T")


class PengScatteringFactor(AbstractFourierOperator, strict=True):
    a: Float[Array, " n"]
    b: Float[Array, " n"]
    b_factor: Float[Array, ""] | None

    def __init__(
        self,
        a: Float[NDArrayLike, " n"],
        b: Float[NDArrayLike, " n"],
        b_factor: FloatLike | None = None,
    ):
        self.a = jnp.asarray(a, dtype=float)
        self.b = jnp.asarray(b, dtype=float)
        self.b_factor = None if b_factor is None else jnp.asarray(b_factor, dtype=float)

    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ):
        q_squared = jnp.sum(frequency_grid**2, axis=-1)
        b_factor = 0.0 if self.b_factor is None else error_if_not_positive(self.b_factor)
        gaussian_fn = lambda _a, _b: _a * jnp.exp(-0.25 * (_b + b_factor) * q_squared)
        return jnp.sum(
            jax.vmap(gaussian_fn)(self.a, error_if_not_positive(self.b)), axis=0
        )


class LobatoScatteringFactor(AbstractFourierOperator, strict=True):
    a: Float[Array, " n"]
    b: Float[Array, " n"]
    b_factor: Float[Array, ""] | None

    def __init__(
        self,
        a: Float[NDArrayLike, " n"],
        b: Float[NDArrayLike, " n"],
        b_factor: FloatLike | None = None,
    ):
        self.a = jnp.asarray(a, dtype=float)
        self.b = jnp.asarray(b, dtype=float)
        self.b_factor = None if b_factor is None else jnp.asarray(b_factor, dtype=float)

    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ):
        q_squared = jnp.sum(frequency_grid**2, axis=-1)
        hydrogenic_fn = (
            lambda _a, _b: _a * (2 + _b * q_squared) / (1 + _b * q_squared) ** 2
        )
        scattering_factor = jnp.sum(jax.vmap(hydrogenic_fn)(self.a, self.b), axis=0)
        if self.b_factor is not None:
            scattering_factor *= jnp.exp(
                -0.25 * error_if_not_positive(self.b_factor) * q_squared
            )
        return scattering_factor


class IndependentAtomVolume(AbstractAtomVolume, strict=True):
    """A representation of a volume that accepts an array of
    atom positions and an electron scattering factor for these
    atoms.

    !!! example "A Gaussian at each atom"
        ```python
        import cryojax.simulator as cxs
        import cryojax.ndimage as im

        positions = ... # load atom positions
        b_factor = ...  # ... and a B-factor
        volume = cxs.IndependentAtomVolume(
            positions=positions, scattering_factors=im.FourierGaussian(b_factor=b_factor)
        )
        ```

    The arguments `positions` and `scattering_factors` may also be
    pytrees of arrays and scattering factors, where each tree leaf represents
    a different atom type.

    !!! example "Multiple atom types"
        ```python
        import cryojax.simulator as cxs
        import cryojax.ndimage as im

        positions_1, positions_2 = ...
        b_factor_1, b_factor_2 = ...
        volume = cxs.IndependentAtomVolume(
            positions=(positions_1, positions_2),
            scattering_factors=(im.FourierGaussian(b_factor=b_factor_1), im.FourierGaussian(b_factor=b_factor_2))
        )
        ```

    See [`cryojax.simulator.IndependentAtomVolume.from_tabulated_parameters`][] for
    loading a volume from tabulated electron scattering factors.
    """  # noqa: E501

    positions: PyTree[Float[Array, "_ 3"]]
    scattering_factors: PyTree[AbstractFourierOperator]

    rotation_convention: ClassVar[Literal["object"]] = "object"

    def __init__(
        self,
        positions: PyTree[Float[NDArrayLike, "_ 3"], "T"],
        scattering_factors: PyTree[AbstractFourierOperator, "T"],
    ):
        """**Arguments:**

        - `positions`:
            A pytree of atom positions.
        - `scattering_factors`:
            A pytree of scattering factors with the same tree structure
            as `positions`, where each leaf is a
            [`cryojax.ndimage.AbstractFourierOperator`][].
        """
        if jax.tree.structure(positions) != jax.tree.structure(
            scattering_factors,
            is_leaf=lambda x: isinstance(x, AbstractFourierOperator),
        ):
            raise ValueError(
                "When instantiating an `IndependentAtomVolume`, found "
                "that the pytree structures of `positions` and "
                "`scattering_factors` were not equal."
            )
        self.positions = jax.tree.map(lambda x: jnp.asarray(x, dtype=float), positions)
        self.scattering_factors = scattering_factors

    @override
    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new potential with rotated `positions`."""
        rotate_fn = lambda pos: pose.rotate_coordinates(pos, inverse=inverse)
        return eqx.tree_at(
            lambda x: x.positions,
            self,
            jax.tree.map(rotate_fn, self.positions),
        )

    @override
    def translate_to_pose(self, pose: AbstractPose) -> Self:
        """Return a new potential with translated `positions`."""
        offset_in_angstroms = pose.offset_in_angstroms
        if pose.offset_z_in_angstroms is None:
            offset_in_angstroms = jnp.concatenate(
                (offset_in_angstroms, jnp.atleast_1d(0.0))
            )
        translate_fn = lambda pos: pos + offset_in_angstroms
        return eqx.tree_at(
            lambda x: x.positions,
            self,
            jax.tree.map(translate_fn, self.positions),
        )

    @classmethod
    def from_tabulated_parameters(
        cls,
        positions_by_element: tuple[Float[NDArrayLike, "_ 3"], ...],
        parameters: PengScatteringFactorParameters | LobatoScatteringFactorParameters,
        *,
        b_factor_by_element: FloatLike | tuple[FloatLike, ...] | None = None,
    ) -> Self:
        def make_scattering_factor(a, b, b_factor):
            if isinstance(parameters, PengScatteringFactorParameters):
                return PengScatteringFactor(a, b, b_factor)
            elif isinstance(parameters, LobatoScatteringFactorParameters):
                return LobatoScatteringFactor(a, b, b_factor)
            else:
                raise ValueError(
                    "Unrecognized argument `parameters` when "
                    "calling `IndependentAtomVolume.from_tabulated_parameters`. "
                    "Should be either `cryojax.constants.PengScatteringFactorParameters` "
                    "or `cryojax.constants.LobatoScatteringFactorParameters`, but got "
                    f"type {parameters.__class__.__name__}."
                )

        n_elements = len(positions_by_element)
        a, b = parameters.a, parameters.b
        if a.shape[0] != n_elements or b.shape[0] != n_elements:
            raise ValueError(
                "When constructing an `IndependentAtomVolume` via "
                "`from_tabulated_parameters`, found that "
                "`parameters.a.shape[0] != len(positions_by_element)` "
                "or `parameters.b.shape[0] != len(positions_by_element)`. "
                "Make sure that `a` and `b` correspond to the element types "
                "in `positions_by_element.`"
            )
        if b_factor_by_element is not None:
            if isinstance(b_factor_by_element, Sequence):
                if len(b_factor_by_element) != n_elements:
                    raise ValueError(
                        "When constructing an `IndependentAtomVolume` via "
                        "`from_tabulated_parameters`, found that "
                        "`len(b_factor_by_element) != len(positions_by_element)`. "
                        "Make sure that `b_factor_by_element` is a tuple with "
                        "length matching the number of atom types."
                    )
            else:
                b_factor_by_element = tuple(
                    b_factor_by_element for _ in range(n_elements)
                )
            scattering_factors_by_element = tuple(
                make_scattering_factor(a_i, b_i, b_factor)
                for a_i, b_i, b_factor in zip(
                    parameters.a, parameters.b, b_factor_by_element
                )
            )
        else:
            scattering_factors_by_element = tuple(
                make_scattering_factor(a_i, b_i, b_factor=None)
                for a_i, b_i in zip(parameters.a, parameters.b)
            )
        return cls(positions_by_element, scattering_factors_by_element)


class FFTAtomRenderFn(AbstractVolumeRenderFn[IndependentAtomVolume], strict=True):
    """Render a voxel grid using non-uniform FFTs and convolution."""

    shape: tuple[int, int, int]
    voxel_size: Float[Array, ""]
    frequency_grid: Float[Array, "_ _ _ 3"] | None
    sampling_mode: Literal["average", "point"]
    eps: float
    opts: Any

    def __init__(
        self,
        shape: tuple[int, int, int],
        voxel_size: FloatLike,
        *,
        frequency_grid: Float[Array, "_ _ _ 3"] | None = None,
        sampling_mode: Literal["average", "point"] = "average",
        eps: float = 1e-6,
        opts: Any = None,
    ):
        """**Arguments:**

        - `shape`:
            The shape of the resulting voxel grid.
        - `voxel_size`:
            The voxel size of the resulting voxel grid.
        - `frequency_grid`:
            An optional frequency grid for rendering the
            volume. If `None`, compute on the fly. The grid
            should be in inverse angstroms and have the zero
            frequency component in the center, i.e.

            ```python
            frequency_grid = jnp.fft.fftshift(
                make_frequency_grid(shape, voxel_size, outputs_rfft=False),
                axes=(0, 1, 2),
            )
            ```

        - `sampling_mode`:
            If `'average'`, convolve with a box function to sample the
            projected volume at a pixel to be the average value of the
            underlying continuous function. If `'point'`, the volume at
            a pixel will be point sampled.
        - `eps`:
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        - `opts`:
            A `jax_finufft.options.Opts` or `jax_finufft.options.NestedOpts`
            dataclass.
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        """
        if sampling_mode not in ["average", "point"]:
            raise ValueError(
                "`sampling_mode` in `FFTAtomRenderFn` "
                "must be either 'average' for averaging within a "
                "pixel or 'point' for point sampling. Got "
                f"`sampling_mode = {sampling_mode}`."
            )
        self.shape = shape
        self.voxel_size = jnp.asarray(voxel_size, dtype=float)
        self.frequency_grid = frequency_grid
        self.sampling_mode = sampling_mode
        self.eps = eps
        self.opts = opts

    @override
    def __call__(
        self,
        volume_representation: IndependentAtomVolume,
        *,
        outputs_real_space: bool = True,
        outputs_rfft: bool = False,
        fftshifted: bool = False,
    ) -> VoxelArray:
        """**Arguments:**

        - `volume_representation`:
            The `GaussianMixtureVolume`.
        - `outputs_real_space`:
            If `True`, return a voxel grid in real-space.
        - `outputs_rfft`:
            If `True`, return a fourier-space voxel grid transformed with
            `cryojax.ndimage.rfftn`. Otherwise, use `fftn`. Does nothing
            if `outputs_real_space = True`.
        - `fftshifted`:
            If `True`, return a fourier-space voxel grid with the zero
            frequency component in the center of the grid via
            `jax.numpy.fft.fftshift`. Otherwise, the zero frequency
            component is in the corner. Does nothing if
            `outputs_real_space = True`.
        """
        voxel_size = error_if_not_positive(self.voxel_size)
        if self.frequency_grid is None:
            frequency_grid = jnp.fft.fftshift(
                make_frequency_grid(self.shape, voxel_size, outputs_rfftfreqs=False),
                axes=(0, 1, 2),
            )
        else:
            frequency_grid = self.frequency_grid
        proj_kernel = lambda pos, kernel: _render_with_nufft(
            self.shape,
            voxel_size,
            pos,
            kernel,
            frequency_grid,
            eps=self.eps,
            opts=self.opts,
        )
        # Compute projection over atom types
        fourier_voxel_grid = jax.tree.reduce(
            lambda x, y: x + y,
            jax.tree.map(
                proj_kernel,
                volume_representation.positions,
                volume_representation.scattering_factors,
                is_leaf=lambda x: isinstance(x, AbstractFourierOperator),
            ),
        )
        if self.sampling_mode == "average":
            antialias_fn = FourierSinc(box_width=voxel_size)
            fourier_voxel_grid *= antialias_fn(frequency_grid)

        if outputs_real_space:
            return ifftn(jnp.fft.ifftshift(fourier_voxel_grid)).real
        else:
            if outputs_rfft:
                fourier_voxel_grid = convert_fftn_to_rfftn(
                    jnp.fft.ifftshift(fourier_voxel_grid), mode="real"
                )
                if fftshifted:
                    return jnp.fft.fftshift(fourier_voxel_grid, axes=(0, 1))
                else:
                    return fourier_voxel_grid
            else:
                if fftshifted:
                    return fourier_voxel_grid
                else:
                    return jnp.fft.ifftshift(fourier_voxel_grid)


class FFTAtomProjection(
    AbstractVolumeIntegrator[IndependentAtomVolume],
    strict=True,
):
    """Integrate atomic parametrization of a volume onto
    the exit plane using non-uniform FFTs plus convolution.
    """

    sampling_mode: Literal["average", "point"]
    upsample_factor: int | None
    shape: tuple[int, int] | None
    eps: float
    opts: Any

    outputs_ewald_sphere: ClassVar[bool] = False

    def __init__(
        self,
        *,
        sampling_mode: Literal["average", "point"] = "average",
        upsample_factor: int | None = None,
        shape: tuple[int, int] | None = None,
        eps: float = 1e-6,
        opts: Any = None,
    ):
        """**Arguments:**

        - `sampling_mode`:
            If `'average'`, convolve with a box function to sample the
            projected volume at a pixel to be the average value of the
            underlying continuous function. If `'point'`, the volume at
            a pixel will be point sampled.
        - `upsample_factor`:
            If provided, first compute an upsampled version of the
            image at pixel size `image_config.pixel_size / upsample_factor`.
            Then, downsample with `cryojax.ndimage.block_reduce_downsample`
            to locally average to the correct pixel size. This is useful
            for reducing aliasing.
        - `shape`:
            If given, first compute the image at `shape`, then
            pad or crop to `image_config.padded_shape`.
        - `eps`:
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        - `opts`:
            A `jax_finufft.options.Opts` or `jax_finufft.options.NestedOpts`
            dataclass.
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        """
        if jnufft is None:
            raise RuntimeError(
                "Tried to use the `FFTAtomProjection` "
                "class, but `jax-finufft` is not installed. "
                "See https://github.com/flatironinstitute/jax-finufft "
                "for installation instructions."
            ) from JAX_FINUFFT_IMPORT_ERROR
        if sampling_mode not in ["average", "point"]:
            raise ValueError(
                "`sampling_mode` in `FFTAtomProjection` "
                "must be either 'average' for averaging within a "
                "pixel or 'point' for point sampling. Got "
                f"`sampling_mode = {sampling_mode}`."
            )
        self.sampling_mode = sampling_mode
        self.upsample_factor = upsample_factor
        self.shape = shape
        self.eps = eps
        self.opts = opts

    @override
    def integrate(
        self,
        volume_representation: IndependentAtomVolume,
        image_config: AbstractImageConfig,
        outputs_real_space: bool = False,
    ) -> ProjectionArray:
        """Compute a projection from scattering factors per atom type
        from the `IndependentAtomVolume`.

        **Arguments:**

        - `volume_representation`:
            The volume representation.
        - `image_config`:
            The configuration of the resulting image.
        - `outputs_real_space`:
            If `True`, return the image in real space. Otherwise,
            return in fourier.

        **Returns:**

        The volume projection in real or Fourier space at the
        `AbstractImageConfig.padded_shape` and the `image_config.pixel_size`.
        """  # noqa: E501
        u = self.upsample_factor
        pixel_size = image_config.pixel_size
        shape = image_config.padded_shape if self.shape is None else self.shape
        if u is None:
            shape_u, pixel_size_u = shape, pixel_size
        else:
            shape_u, pixel_size_u = (u * shape[0], u * shape[1]), pixel_size / u
        if shape_u == image_config.padded_shape:
            frequency_grid = image_config.get_frequency_grid(
                padding=True, physical=True, full=True
            )
        else:
            frequency_grid = make_frequency_grid(
                shape_u, pixel_size_u, outputs_rfftfreqs=False
            )
        frequency_grid = jnp.fft.fftshift(frequency_grid, axes=(0, 1))
        proj_kernel = lambda pos, kernel: _project_with_nufft(
            shape_u,
            pixel_size_u,
            pos,
            kernel,
            frequency_grid,
            eps=self.eps,
            opts=self.opts,
        )
        # Compute projection over atom types
        fourier_projection = jax.tree.reduce(
            lambda x, y: x + y,
            jax.tree.map(
                proj_kernel,
                volume_representation.positions,
                volume_representation.scattering_factors,
                is_leaf=lambda x: isinstance(x, AbstractFourierOperator),
            ),
        )
        # Apply anti-aliasing filter
        if self.sampling_mode == "average":
            antialias_fn = FourierSinc(box_width=pixel_size_u)
            fourier_projection *= antialias_fn(frequency_grid)
        # If upsample factor is even, need subpixel center correction
        if u is not None and u % 2 == 0:
            fourier_projection = _apply_subpixel_shift(
                shape,
                fourier_projection,
                pixel_size_u * frequency_grid,
                u,
            )
        # Shift zero frequency component to corner and convert to
        # rfft
        fourier_projection = convert_fftn_to_rfftn(
            jnp.fft.ifftshift(fourier_projection), mode="real"
        )
        if self.shape is None:
            if u is None:
                return (
                    irfftn(fourier_projection, s=shape)
                    if outputs_real_space
                    else fourier_projection
                )
            else:
                projection = _block_average(irfftn(fourier_projection, s=shape_u), u)
                return projection if outputs_real_space else rfftn(projection)
        else:
            projection = irfftn(fourier_projection, s=shape_u)
            if u is not None:
                projection = _block_average(projection, u)
            projection = resize_with_crop_or_pad(projection, image_config.padded_shape)
            return projection if outputs_real_space else rfftn(projection)


def _render_with_nufft(shape, ps, pos, kernel, freqs, eps=1e-6, opts=None):
    assert jnufft is not None
    # Get x and y coordinates
    # Normalize coordinates betweeen -pi and pi
    nz, ny, nx = shape
    box_xyz = ps * jnp.asarray((nx, ny, nz))
    pos_periodic = 2 * jnp.pi * pos / box_xyz
    # Unpack
    x, y, z = pos_periodic[:, 0], pos_periodic[:, 1], pos_periodic[:, 2]
    n_atoms = x.size
    volume_element = ps**3
    # Compute
    fourier_projection = kernel(freqs) * (
        jnufft.nufft1(
            shape,
            jnp.full((n_atoms,), 1.0 + 0.0j),
            z,
            y,
            x,
            eps=eps,
            opts=opts,
            iflag=-1,
        )
        / volume_element
    )

    return fourier_projection


def _project_with_nufft(shape, ps, pos, kernel, freqs, eps=1e-6, opts=None):
    assert jnufft is not None
    # Get x and y coordinates
    positions_xy = pos[:, :2]
    # Normalize coordinates betweeen -pi and pi
    ny, nx = shape
    box_xy = ps * jnp.asarray((nx, ny))
    positions_periodic = 2 * jnp.pi * positions_xy / box_xy
    # Unpack
    x, y = positions_periodic[:, 0], positions_periodic[:, 1]
    n_atoms = x.size
    area_element = ps**2
    # Compute
    fourier_projection = kernel(freqs) * (
        jnufft.nufft1(
            shape,
            jnp.full((n_atoms,), 1.0 + 0.0j),
            y,
            x,
            eps=eps,
            opts=opts,
            iflag=-1,
        )
        / area_element
    )

    return fourier_projection


def _block_average(x, factor):
    return (
        block_reduce_downsample(x, factor, jax.lax.add, center_correct=False)
        / factor**x.ndim
    )


def _apply_subpixel_shift(target_shape, fourier_image, frequency_grid, k):
    if len(set(target_shape)) > 1:
        raise NotImplementedError(
            "Even `upsample_factor` and non-square image shape not "
            f"supported in `FFTAtomProjection`. Got `upsample_factor = {k}` "
            f"and `shape = {target_shape}`."
        )
    dim = target_shape[0]
    if dim % 2 == 0:
        shift = jnp.full((2,), (k - 1) / 2)
    else:
        shift = jnp.full((2,), -0.5)
    return (
        jnp.exp(-1.0j * (2 * jnp.pi * jnp.matmul(frequency_grid, shift))) * fourier_image
    )



import warnings
from collections.abc import Callable
from typing import Any, ClassVar, Literal, TypedDict
from typing_extensions import Self, override

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Array, Float, PyTree

from ..._internal import error_if_not_positive
from ...constants import (
    PengScatteringFactorParameters,
    b_factor_to_variance,
    variance_to_b_factor,
)
from ...jax_util import FloatLike, NDArrayLike
from ...ndimage import fftn, make_1d_coordinate_grid, resize_with_crop_or_pad, rfftn
from .._image_config import AbstractImageConfig
from .._pose import AbstractPose
from .base_volume import (
    AbstractAtomVolume,
    AbstractVolumeIntegrator,
    AbstractVolumeRenderFn,
    ProjectionArray,
    VoxelArray,
)


class BatchOptions(TypedDict):
    batch_size: int
    n_batches: int


class GaussianMixtureVolume(AbstractAtomVolume, strict=True):
    r"""A representation of a volume as a mixture of
    gaussians, with multiple gaussians used per position.

    The convention of allowing multiple gaussians per position
    follows "Robust Parameterization of Elastic and Absorptive
    Electron Atomic Scattering Factors" by Peng et al. (1996). The
    $a$ and $b$ parameters in this work correspond to
    `amplitudes = a` and `variances = b / 8\pi^2`.

    !!! info
        Use the following to load a `GaussianMixtureVolume`
        from these tabulated electron scattering factors.

        ```python
        from cryojax.constants import PengScatteringFactorParameters
        from cryojax.io import read_atoms_from_pdb
        from cryojax.simulator import GaussianMixtureVolume

        # Load positions of atoms and one-hot encoded atom names
        atom_positions, atom_types = read_atoms_from_pdb(...)
        parameters = PengScatteringFactorParameters(atom_types)
        potential = GaussianMixtureVolume.from_tabulated_parameters(
            atom_positions, parameters
        )
        ```
    """

    positions: Float[Array, "n_positions 3"]
    amplitudes: Float[Array, "n_positions n_gaussians"]
    variances: Float[Array, " n_positions n_gaussians"]

    rotation_convention: ClassVar[Literal["object"]] = "object"

    def __init__(
        self,
        positions: Float[NDArrayLike, "n_positions 3"],
        amplitudes: (
            float
            | Float[NDArrayLike, ""]
            | Float[NDArrayLike, " n_positions"]
            | Float[NDArrayLike, "n_positions n_gaussians"]
        ),
        variances: (
            float
            | Float[NDArrayLike, ""]
            | Float[NDArrayLike, " n_positions"]
            | Float[NDArrayLike, "n_positions n_gaussians"]
        ),
    ):
        """**Arguments:**

        - `positions`:
            The coordinates of the gaussians in units of angstroms.
        - `amplitudes`:
            The amplitude for each gaussian.
            To simulate in physical units of a scattering potential,
            this should have units of angstroms.
        - `variances`:
            The variance for each gaussian. This has units of angstroms
            squared.
        """
        n_positions = positions.shape[0]
        if isinstance(amplitudes, NDArrayLike):
            if amplitudes.ndim == 2:
                n_gaussians = amplitudes.shape[-1]
            elif amplitudes.ndim == 1:
                n_gaussians = 1
                amplitudes = amplitudes[:, None]
            elif amplitudes.ndim == 0:
                n_gaussians = 1
                amplitudes = amplitudes[None, None]
            else:
                raise ValueError(
                    "Passed `amplitudes` to `GaussianMixtureVolume` "
                    f"with shape {amplitudes.shape}, but must be of "
                    "shape `()`, `(n_positions,)`, or "
                    "`(n_positions, n_gaussians)`."
                )
        else:
            n_gaussians = 1
        if isinstance(variances, NDArrayLike):
            if variances.ndim == 2:
                n_gaussians = variances.shape[-1]
            elif variances.ndim == 1:
                variances = variances[:, None]
            elif variances.ndim == 0:
                variances = variances[None, None]
            else:
                raise ValueError(
                    "Passed `variances` to `GaussianMixtureVolume` "
                    f"with shape {variances.shape}, but must be of "
                    "shape `()`, `(n_positions,)`, or "
                    "`(n_positions, n_gaussians)`."
                )

        self.positions = jnp.asarray(positions, dtype=float)
        self.amplitudes = jnp.broadcast_to(
            jnp.asarray(amplitudes, dtype=float), (n_positions, n_gaussians)
        )
        self.variances = jnp.broadcast_to(
            jnp.asarray(variances, dtype=float), (n_positions, n_gaussians)
        )

    def __check_init__(self):
        if not (
            self.positions.shape[0] == self.amplitudes.shape[0] == self.variances.shape[0]
        ):
            raise ValueError(
                "The number of positions in `GaussianMixtureVolume` was "
                f"{self.positions.shape[0]}, but `amplitudes` shape was "
                f"{self.amplitudes.shape} and `variances` shape was "
                f"{self.variances.shape}. The first dimension must be equal "
                "to the number of positions."
            )
        if not (self.amplitudes.shape == self.variances.shape):
            raise ValueError(
                "In `GaussianMixtureVolume`, `amplitudes` and "
                f"`variances` shape must be equal. Found shapes "
                f"{self.amplitudes.shape} and {self.variances.shape}, "
                "respectively."
            )

    @classmethod
    def from_tabulated_parameters(
        cls,
        atom_positions: Float[NDArrayLike, "n_atoms 3"],
        parameters: PengScatteringFactorParameters,
        extra_b_factors: FloatLike | Float[NDArrayLike, " n_atoms"] | None = None,
    ) -> Self:
        """Initialize a `GaussianMixtureVolume` from tabulated electron
        scattering factor parameters (Peng et al. 1996). This treats
        the scattering potential as a mixture of five gaussians
        per atom.

        **References:**

        - Peng, L-M. "Electron atomic scattering factors and scattering potentials of crystals."
            Micron 30.6 (1999): 625-648.
        - Peng, L-M., et al. "Robust parameterization of elastic and absorptive electron atomic
            scattering factors." Acta Crystallographica Section A: Foundations of Crystallography
            52.2 (1996): 257-276.

        **Arguments:**

        - `atom_positions`:
            The coordinates of the atoms in units of angstroms.
        - `parameters`:
            A pytree for the scattering factor parameters from
            Peng et al. (1996).
        - `extra_b_factors`:
            Additional per-atom B-factors that are added to
            the values in `scattering_parameters.b`.
        """  # noqa: E501
        amplitudes = jnp.asarray(parameters.a, dtype=float)
        b_factors = jnp.asarray(parameters.b, dtype=float)
        if extra_b_factors is not None:
            extra_b_factors = jnp.asarray(extra_b_factors, dtype=float)
            if extra_b_factors.ndim == 1:
                extra_b_factors = extra_b_factors[:, None]
            b_factors += extra_b_factors
        return cls(atom_positions, amplitudes, b_factor_to_variance(b_factors))

    @override
    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new potential with rotated `positions`."""
        return eqx.tree_at(
            lambda d: d.positions,
            self,
            pose.rotate_coordinates(self.positions, inverse=inverse),
        )

    @override
    def translate_to_pose(self, pose: AbstractPose) -> Self:
        """Return a new potential with rotated `positions`."""
        offset_in_angstroms = pose.offset_in_angstroms
        if pose.offset_z_in_angstroms is None:
            offset_in_angstroms = jnp.concatenate(
                (offset_in_angstroms, jnp.atleast_1d(0.0))
            )
        return eqx.tree_at(
            lambda d: d.positions, self, self.positions + offset_in_angstroms
        )

    def to_real_voxel_grid(
        self,
        shape: tuple[int, int, int],
        voxel_size: Float[NDArrayLike, ""] | float,
        *,
        batch_options: dict[str, Any] = {},
    ) -> Float[Array, "{shape[0]} {shape[1]} {shape[2]}"]:
        warnings.warn(
            "'GaussianMixtureVolume.to_real_voxel_grid' is deprecated "
            "and will be removed in cryoJAX 0.6.0. Instead, use "
            "`cryojax.simulator.GaussianMixtureRenderFn`.",
            category=FutureWarning,
            stacklevel=2,
        )
        return _gaussians_to_real_voxels(
            shape,
            jnp.asarray(voxel_size, dtype=float),
            self.positions,
            self.amplitudes,
            variance_to_b_factor(self.variances),
            **batch_options,
        )


class GaussianMixtureProjection(
    AbstractVolumeIntegrator[GaussianMixtureVolume],
    strict=True,
):
    sampling_mode: Literal["average", "point"]
    shape: tuple[int, int] | None
    n_batches: int

    outputs_ewald_sphere: ClassVar[bool] = False

    def __init__(
        self,
        *,
        upsampling_factor: int | None = None,
        shape: tuple[int, int] | None = None,
        sampling_mode: Literal["average", "point"] = "average",
        use_error_functions: bool = False,
        n_batches: int = 1,
    ):
        """**Arguments:**

        - `shape`:
            The shape of the plane on which projections are computed before padding or
            cropping to the `AbstractImageConfig.padded_shape`. This argument is particularly
            useful if the `AbstractImageConfig.padded_shape` is much larger than the protein.
        - `sampling_mode`:
            If `'average'`, use error functions to sample the projected volume at
            a pixel to be the average value using gaussian
            integrals. If `'point'`, the volume at a pixel will
            be evaluated by evaluating the gaussian at a point.
        - `n_batches`:
            The number of batches over groups of positions
            used to evaluate the projection. By default, `n_batches = 1`,
            which computes a projection for all positions at once.
            This is useful to decrease GPU memory usage.
        """  # noqa: E501
        if use_error_functions:
            warnings.warn(
                "`use_error_functions` in `GaussianMixtureProjection` has "
                "been deprecated and will be removed in cryoJAX 0.6.0. "
                "This has been renamed to `sampling_mode = 'average'`.",
                category=FutureWarning,
                stacklevel=2,
            )
        if upsampling_factor is not None:
            raise ValueError(
                "`upsampling_factor` in `GaussianMixtureProjection` "
                "has been deprecated as of cryoJAX 0.5.1. The "
                "functionality this implemented was not as intended."
            )
        if sampling_mode not in ["average", "point"]:
            raise ValueError(
                "`sampling_mode` in `GaussianMixtureProjection` "
                "must be either 'average' for averaging within a "
                "pixel or 'point' for point sampling. Got "
                f"`sampling_mode = {sampling_mode}`."
            )
        self.shape = shape
        self.sampling_mode = sampling_mode
        self.n_batches = n_batches

    @override
    def integrate(
        self,
        volume_representation: GaussianMixtureVolume,
        image_config: AbstractImageConfig,
        outputs_real_space: bool = False,
    ) -> ProjectionArray:
        """Compute a projection from gaussians.

        **Arguments:**

        - `volume_representation`:
            The volume representation.
        - `image_config`:
            The image configuration.
        - `outputs_real_space`:
            If `True`, return the image in real space. Otherwise,
            return in Fourier.

        **Returns:**

        The volume projection in real or Fourier space at the
        `AbstractImageConfig.padded_shape` and the `image_config.pixel_size`.
        """  # noqa: E501
        # Grab the image configuration
        shape = image_config.padded_shape if self.shape is None else self.shape
        pixel_size = image_config.pixel_size
        # Grab the gaussian amplitudes and widths
        positions = volume_representation.positions
        amplitudes = volume_representation.amplitudes
        b_factors = variance_to_b_factor(
            error_if_not_positive(volume_representation.variances)
        )
        # Compute the projection
        use_erf = True if self.sampling_mode == "average" else False
        projection_integral = _gaussians_to_projection(
            shape,
            pixel_size,
            positions,
            amplitudes,
            b_factors,
            use_erf,
            self.n_batches,
        )
        if self.shape is None:
            return (
                projection_integral if outputs_real_space else rfftn(projection_integral)
            )
        else:
            projection_integral = resize_with_crop_or_pad(
                projection_integral, image_config.padded_shape
            )
            return (
                projection_integral if outputs_real_space else rfftn(projection_integral)
            )


class GaussianMixtureRenderFn(AbstractVolumeRenderFn[GaussianMixtureVolume], strict=True):
    """Render a voxel grid from the `GaussianMixtureVolume`.

    If `GaussianMixtureVolume` is instantiated from electron scattering
    factors via `from_tabulated_parameters`, this renders an electrostatic
    potential as tabulated in Peng et al. 1996. The elastic electron
    scattering factors defined in this work are

    $$f^{(e)}(\\mathbf{q}) = \\sum\\limits_{i = 1}^5 a_i \\exp(- b_i |\\mathbf{q}|^2),$$

    where $a_i$ is stored as `GaussianMixtureVolume.amplitudes`,
    $b_i / 8 \\pi^2$ are the `GaussianMixtureVolume.variances`, and
    $\\mathbf{q}$ is the scattering vector.

    Under usual scattering approximations (i.e. the first-born approximation),
    the rescaled electrostatic potential energy $U(\\mathbf{r})$ for a given atom type is
    $\\mathcal{F}^{-1}[f^{(e)}(\\boldsymbol{\\xi} / 2)](\\mathbf{r})$, which is computed
    analytically as

    $$U(\\mathbf{r}) = \\sum\\limits_{i = 1}^5 \\frac{a_i}{(2\\pi (b_i / 8 \\pi^2))^{3/2}} \\exp(- \\frac{|\\mathbf{r} - \\mathbf{r}'|^2}{2 (b_i / 8 \\pi^2)}),$$

    where $\\mathbf{r}'$ is the position of the atom. Including an additional B-factor (denoted by
    $B$) gives the expression for the potential
    $U(\\mathbf{r})$ of a single atom type and its fourier transform pair $\\tilde{U}(\\boldsymbol{\\xi}) \\equiv \\mathcal{F}[U](\\boldsymbol{\\xi})$,

    $$U(\\mathbf{r}) = \\sum\\limits_{i = 1}^5 \\frac{a_i}{(2\\pi ((b_i + B) / 8 \\pi^2))^{3/2}} \\exp(- \\frac{|\\mathbf{r} - \\mathbf{r}'|^2}{2 ((b_i + B) / 8 \\pi^2)}),$$

    $$\\tilde{U}(\\boldsymbol{\\xi}) = \\sum\\limits_{i = 1}^5 a_i \\exp(- (b_i + B) |\\boldsymbol{\\xi}|^2 / 4) \\exp(2 \\pi i \\boldsymbol{\\xi}\\cdot\\mathbf{r}'),$$

    where $\\mathbf{q} = \\boldsymbol{\\xi} / 2$ gives the relationship between the wave vector and the
    scattering vector.

    In practice, for a discretization on a grid with voxel size $\\Delta r$ and grid point $\\mathbf{r}_{\\ell}$,
    the potential is evaluated as the average value inside the voxel

    $$U_{\\ell} = \\frac{1}{\\Delta r^3} \\sum\\limits_{i = 1}^5 a_i \\prod\\limits_{j = 1}^3 \\int_{r^{\\ell}_j-\\Delta r/2}^{r^{\\ell}_j+\\Delta r/2} dr_j \\ \\frac{1}{{\\sqrt{2\\pi ((b_i + B) / 8 \\pi^2)}}} \\exp(- \\frac{(r_j - r'_j)^2}{2 ((b_i + B) / 8 \\pi^2)}),$$

    where $j$ indexes the components of the spatial coordinate vector $\\mathbf{r}$. The above expression is evaluated using the error function as

    $$U_{\\ell} = \\frac{1}{(2 \\Delta r)^3} \\sum\\limits_{i = 1}^5 a_i \\prod\\limits_{j = 1}^3 \\textrm{erf}(\\frac{r_j^{\\ell} - r'_j + \\Delta r / 2}{\\sqrt{2 ((b_i + B) / 8\\pi^2)}}) - \\textrm{erf}(\\frac{r_j^{\\ell} - r'_j - \\Delta r / 2}{\\sqrt{2 ((b_i + B) / 8\\pi^2)}}).$$
    """  # noqa: E501

    shape: tuple[int, int, int]
    voxel_size: Float[Array, ""]

    batch_options: BatchOptions

    def __init__(
        self,
        shape: tuple[int, int, int],
        voxel_size: FloatLike,
        *,
        batch_options: dict[str, Any] = {},
    ):
        """**Arguments:**

        - `shape`:
            The shape of the resulting voxel grid.
        - `voxel_size`:
            The voxel size of the resulting voxel grid.
        - `batch_options`:
            Advanced options for controlling batching. This is a dictionary
            with the following keys:
            - "batch_size":
                The number of z-planes to evaluate in parallel with
                `jax.vmap`. By default, `1`.
            - "n_batches":
                The number of iterations used to evaluate the volume,
                where the iteration is taken over groups of atoms.
                This is useful if `batch_size = 1`
                and GPU memory is exhausted. By default, `1`.
        """
        self.shape = shape
        self.voxel_size = jnp.asarray(voxel_size, dtype=float)
        self.batch_options = _dict_to_batch_options(batch_options)

    @override
    def __call__(
        self,
        volume_representation: GaussianMixtureVolume,
        *,
        outputs_real_space: bool = True,
        outputs_rfft: bool = False,
        fftshifted: bool = False,
    ) -> VoxelArray:
        """**Arguments:**

        - `volume_representation`:
            The `GaussianMixtureVolume`.
        - `outputs_real_space`:
            If `True`, return a voxel grid in real-space.
        - `outputs_rfft`:
            If `True`, return a fourier-space voxel grid transformed with
            `cryojax.ndimage.rfftn`. Otherwise, use `fftn`. Does nothing
            if `outputs_real_space = True`.
        - `fftshifted`:
            If `True`, return a fourier-space voxel grid with the zero
            frequency component in the center of the grid via
            `jax.numpy.fft.fftshift`. Otherwise, the zero frequency
            component is in the corner. Does nothing if
            `outputs_real_space = True`.
        """
        real_voxel_grid = _gaussians_to_real_voxels(
            self.shape,
            error_if_not_positive(self.voxel_size),
            volume_representation.positions,
            volume_representation.amplitudes,
            variance_to_b_factor(error_if_not_positive(volume_representation.variances)),
            **self.batch_options,
        )
        if outputs_real_space:
            return real_voxel_grid
        else:
            if outputs_rfft:
                return (
                    jnp.fft.fftshift(rfftn(real_voxel_grid), axes=(0, 1))
                    if fftshifted
                    else rfftn(real_voxel_grid)
                )
            else:
                return (
                    jnp.fft.fftshift(fftn(real_voxel_grid))
                    if fftshifted
                    else fftn(real_voxel_grid)
                )


def _dict_to_batch_options(d):
    batch_size = 1 if "batch_size" not in d else d["batch_size"]
    n_batches = 1 if "n_batches" not in d else d["n_batches"]
    return BatchOptions(batch_size=batch_size, n_batches=n_batches)


#
# Projection
#
def _gaussians_to_projection(
    shape: tuple[int, int],
    pixel_size: Float[Array, ""],
    positions: Float[Array, "n_positions 3"],
    a: Float[Array, "n_positions n_gaussians_per_position"],
    b: Float[Array, "n_positions n_gaussians_per_position"],
    use_erf: bool,
    n_batches: int,
) -> Float[Array, "dim_y dim_x"]:
    # Make the grid on which to evaluate the result
    grid_x = make_1d_coordinate_grid(shape[1], pixel_size)
    grid_y = make_1d_coordinate_grid(shape[0], pixel_size)
    # Get function and pytree to compute volume over a batch of positions
    xs = (positions, a, b)
    kernel_fn = lambda xs: _gaussians_to_projection_kernel(
        grid_x,
        grid_y,
        pixel_size,
        xs[0],
        xs[1],
        xs[2],
        use_erf,
    )
    # Compute projection with a call to `jax.lax.map` in batches
    if n_batches > positions.shape[0]:
        raise ValueError(
            "The `n_batches` when computing a projection must "
            "be an integer less than or equal to the number of positions, "
            f"which is equal to {positions.shape[0]}. Got "
            f"`n_batches = {n_batches}`."
        )
    elif n_batches == 1:
        projection = kernel_fn(xs)
    elif n_batches > 1:
        projection = jnp.sum(
            _batched_map_with_contraction(kernel_fn, xs, n_batches),
            axis=0,
        )
    else:
        raise ValueError(
            "The `n_batches` argument for `GaussianMixtureProjection` must be an "
            "integer greater than or equal to 1."
        )
    return projection


def _gaussians_to_projection_kernel(
    grid_x: Float[Array, " dim_x"],
    grid_y: Float[Array, " dim_y"],
    pixel_size: Float[Array, ""],
    positions: Float[Array, "n_positions 3"],
    a: Float[Array, "n_positions n_gaussians_per_position"],
    b: Float[Array, "n_positions n_gaussians_per_position"],
    use_erf: bool,
) -> Float[Array, "dim_y dim_x"]:
    # Evaluate 1D gaussian integrals for each of x, y, and z dimensions

    if use_erf:
        gaussians_times_prefactor_x, gaussians_y = _evaluate_gaussian_integrals_2d(
            grid_x, grid_y, positions, a, b, pixel_size
        )
    else:
        gaussians_times_prefactor_x, gaussians_y = _evaluate_gaussians_2d(
            grid_x, grid_y, positions, a, b
        )
    projection = _evaluate_multivariate_gaussian_2d(
        gaussians_times_prefactor_x, gaussians_y
    )

    return projection


def _evaluate_multivariate_gaussian_2d(
    gaussians_per_interval_per_position_x: Float[
        Array, "dim_x n_positions n_gaussians_per_position"
    ],
    gaussians_per_interval_per_position_y: Float[
        Array, "dim_y n_positions n_gaussians_per_position"
    ],
) -> Float[Array, "dim_y dim_x"]:
    # Prepare matrices with dimensions of the number of positions and the number of grid
    # points. There are as many matrices as number of gaussians per position
    gauss_x = jnp.transpose(gaussians_per_interval_per_position_x, (2, 1, 0))
    gauss_y = jnp.transpose(gaussians_per_interval_per_position_y, (2, 0, 1))
    # Compute matrix multiplication then sum over the number of gaussians per position
    return jnp.sum(jnp.matmul(gauss_y, gauss_x), axis=0)


def _evaluate_gaussian_integrals_2d(
    grid_x: Float[Array, " dim_x"],
    grid_y: Float[Array, " dim_y"],
    positions: Float[Array, "n_positions 3"],
    a: Float[Array, "n_positions n_gaussians_per_position"],
    b: Float[Array, "n_positions n_gaussians_per_position"],
    pixel_size: Float[Array, ""],
) -> tuple[
    Float[Array, "dim_x n_positions n_gaussians_per_position"],
    Float[Array, "dim_y n_positions n_gaussians_per_position"],
]:
    """Evaluate 1D averaged gaussians in x, y, and z dimensions
    for each position and each gaussian per position.
    """
    # Define function to compute integrals for each dimension
    scaling = 2 * jnp.pi / jnp.sqrt(b)
    integration_kernel = lambda delta: (
        jsp.special.erf(scaling[None, :, :] * (delta + pixel_size)[:, :, None])
        - jsp.special.erf(scaling[None, :, :] * delta[:, :, None])
    )
    # Compute outer product of left edge of grid points minus positions
    left_edge_grid_x, left_edge_grid_y = (
        grid_x - pixel_size / 2,
        grid_y - pixel_size / 2,
    )
    delta_x, delta_y = (
        left_edge_grid_x[:, None] - positions[:, 0],
        left_edge_grid_y[:, None] - positions[:, 1],
    )
    # Compute gaussian integrals for each grid point, each position, and
    # each gaussian per position
    gauss_x, gauss_y = (integration_kernel(delta_x), integration_kernel(delta_y))
    # Compute the prefactors for each position and each gaussian per position
    # for the volume
    prefactor = a / (2 * pixel_size) ** 2
    # Multiply the prefactor onto one of the gaussians for efficiency
    return prefactor * gauss_x, gauss_y


def _evaluate_gaussians_2d(
    grid_x: Float[Array, " x_dim"],
    grid_y: Float[Array, " y_dim"],
    positions: Float[Array, "n_positions 3"],
    a: Float[Array, "n_positions n_gaussians_per_position"],
    b: Float[Array, "n_positions n_gaussians_per_position"],
) -> tuple[
    Float[Array, "dim_x n_positions n_gaussians_per_position"],
    Float[Array, "dim_y n_positions n_gaussians_per_position"],
]:
    b_inverse = 4.0 * jnp.pi / b
    gauss_x = jnp.exp(
        -jnp.pi
        * b_inverse[None, :, :]
        * ((grid_x[:, None] - positions.T[0, :]) ** 2)[:, :, None]
    )
    gauss_y = jnp.exp(
        -jnp.pi
        * b_inverse[None, :, :]
        * ((grid_y[:, None] - positions.T[1, :]) ** 2)[:, :, None]
    )
    prefactor = a[None, :, :] * b_inverse[None, :, :]

    return prefactor * gauss_x, gauss_y


#
# Voxel rendering
#
@eqx.filter_jit
def _gaussians_to_real_voxels(
    shape: tuple[int, int, int],
    voxel_size: Float[Array, ""],
    positions: Float[Array, "n_positions 3"],
    amplitudes: Float[Array, "n_positions n_gaussians_per_position"],
    b_factors: Float[Array, "n_positions n_gaussians_per_position"],
    *,
    batch_size: int = 1,
    n_batches: int = 1,
) -> Float[Array, "{shape[0]} {shape[1]} {shape[2]}"]:
    # Make coordinate systems for each of x, y, and z dimensions
    z_dim, y_dim, x_dim = shape
    grid_x, grid_y, grid_z = [
        make_1d_coordinate_grid(dim, voxel_size) for dim in [x_dim, y_dim, z_dim]
    ]
    # Get function to compute potential over a batch of positions
    render_fn = lambda xs: _gaussians_to_real_voxels_kernel(
        grid_x,
        grid_y,
        grid_z,
        voxel_size,
        xs[0],
        xs[1],
        xs[2],
        batch_size,
    )
    if n_batches > positions.shape[0]:
        raise ValueError(
            "The `n_batches` when building a voxel grid must "
            "be an integer less than or equal to the number of positions, "
            f"which is equal to {positions.shape[0]}. Got "
            f"`n_batches = {n_batches}`."
        )
    elif n_batches == 1:
        real_voxel_grid = render_fn((positions, amplitudes, b_factors))
    elif n_batches > 1:
        real_voxel_grid = jnp.sum(
            _batched_map_with_n_batches(
                render_fn,
                (positions, amplitudes, b_factors),
                n_batches=n_batches,
                is_batch_axis_contracted=True,
            ),
            axis=0,
        )
    else:
        raise ValueError(
            "The `n_batches` when building a voxel grid must be an "
            "integer greater than or equal to 1."
        )

    return real_voxel_grid


def _gaussians_to_real_voxels_kernel(
    grid_x: Float[Array, " dim_x"],
    grid_y: Float[Array, " dim_y"],
    grid_z: Float[Array, " dim_z"],
    voxel_size: Float[Array, ""],
    positions: Float[Array, "n_positions_in_batch 3"],
    amplitudes: Float[Array, "n_positions_in_batch n_gaussians_per_position"],
    b_factors: Float[Array, "n_positions_in_batch n_gaussians_per_position"],
    batch_size: int,
) -> Float[Array, "dim_z dim_y dim_x"]:
    # Evaluate 1D gaussian integrals for each of x, y, and z dimensions
    (
        gaussian_integrals_times_prefactor_per_interval_per_position_x,
        gaussian_integrals_per_interval_per_position_y,
        gaussian_integrals_per_interval_per_position_z,
    ) = _evaluate_gaussian_integrals_3d(
        grid_x, grid_y, grid_z, positions, amplitudes, b_factors, voxel_size
    )
    # Get function to compute voxel grid at a single z-plane
    render_at_z_plane = (
        lambda gaussian_integrals_per_position_z: _evaluate_multivariate_gaussian_3d(
            gaussian_integrals_times_prefactor_per_interval_per_position_x,
            gaussian_integrals_per_interval_per_position_y,
            gaussian_integrals_per_position_z,
        )
    )
    # Map over z-planes
    if batch_size > grid_z.size:
        raise ValueError(
            "The `batch_size` when building a voxel grid must be an "
            "integer less than or equal to the z-dimension of the grid, "
            f"which is equal to {grid_z.size}."
        )
    elif batch_size == 1:
        # ... compute the volume iteratively
        real_voxel_grid = jax.lax.map(
            render_at_z_plane, gaussian_integrals_per_interval_per_position_z
        )
    elif batch_size > 1:
        # ... compute the volume by tuning how many z-planes to batch over
        render_at_z_planes = jax.vmap(render_at_z_plane, in_axes=0)
        real_voxel_grid = _batched_map_with_batch_size(
            render_at_z_planes,
            gaussian_integrals_per_interval_per_position_z,
            batch_size=batch_size,
            is_batch_axis_contracted=False,
        )
    else:
        raise ValueError(
            "The `batch_size` when building a voxel grid must be an "
            "integer greater than or equal to 1."
        )

    return real_voxel_grid


def _evaluate_gaussian_integrals_3d(
    grid_x: Float[Array, " dim_x"],
    grid_y: Float[Array, " dim_y"],
    grid_z: Float[Array, " dim_z"],
    positions: Float[Array, "n_positions 3"],
    amplitudes: Float[Array, "n_positions n_gaussians_per_position"],
    b_factors: Float[Array, "n_positions n_gaussians_per_position"],
    voxel_size: Float[Array, ""],
) -> tuple[
    Float[Array, "dim_x n_positions n_gaussians_per_position"],
    Float[Array, "dim_y n_positions n_gaussians_per_position"],
    Float[Array, "dim_z n_positions n_gaussians_per_position"],
]:
    """Evaluate 1D averaged gaussians in x, y, and z dimensions
    for each position and each gaussian per position.
    """
    # Define function to compute integrals for each dimension
    scaling = 2 * jnp.pi / jnp.sqrt(b_factors)
    integration_kernel = lambda delta: (
        jsp.special.erf(scaling[None, :, :] * (delta + voxel_size)[:, :, None])
        - jsp.special.erf(scaling[None, :, :] * delta[:, :, None])
    )
    # Compute outer product of left edge of grid points minus positions
    left_edge_grid_x, left_edge_grid_y, left_edge_grid_z = (
        grid_x - voxel_size / 2,
        grid_y - voxel_size / 2,
        grid_z - voxel_size / 2,
    )
    delta_x, delta_y, delta_z = (
        left_edge_grid_x[:, None] - positions[:, 0],
        left_edge_grid_y[:, None] - positions[:, 1],
        left_edge_grid_z[:, None] - positions[:, 2],
    )
    # Compute gaussian integrals for each grid point, each position, and
    # each gaussian per position
    gauss_x, gauss_y, gauss_z = (
        integration_kernel(delta_x),
        integration_kernel(delta_y),
        integration_kernel(delta_z),
    )
    # Compute the prefactors for each position and each gaussian per position
    # for the potential
    prefactor = amplitudes / (2 * voxel_size) ** 3
    # Multiply the prefactor onto one of the gaussians for efficiency
    return prefactor * gauss_x, gauss_y, gauss_z


def _evaluate_multivariate_gaussian_3d(
    gaussian_integrals_per_interval_per_position_x: Float[
        Array, "dim_x n_positions n_gaussians_per_position"
    ],
    gaussian_integrals_per_interval_per_position_y: Float[
        Array, "dim_y n_positions n_gaussians_per_position"
    ],
    gaussian_integrals_per_position_z: Float[
        Array, "n_positions n_gaussians_per_position"
    ],
) -> Float[Array, "dim_y dim_x"]:
    # Prepare matrices with dimensions of the number of positions and the number of grid
    # points. There are as many matrices as number of gaussians per position
    gauss_x = jnp.transpose(gaussian_integrals_per_interval_per_position_x, (2, 1, 0))
    gauss_yz = jnp.transpose(
        gaussian_integrals_per_interval_per_position_y
        * gaussian_integrals_per_position_z[None, :, :],
        (2, 0, 1),
    )
    # Compute matrix multiplication then sum over the number of gaussians per position
    return jnp.sum(jnp.matmul(gauss_yz, gauss_x), axis=0)


def _batched_map_with_n_batches(
    fun: Callable,
    xs: PyTree[Array],
    n_batches: int,
    is_batch_axis_contracted: bool = False,
):
    batch_dim = jax.tree.leaves(xs)[0].shape[0]
    batch_size = batch_dim // n_batches
    return _batched_map(
        fun, xs, batch_dim, n_batches, batch_size, is_batch_axis_contracted
    )


def _batched_map_with_batch_size(
    fun: Callable,
    xs: PyTree[Array],
    batch_size: int,
    is_batch_axis_contracted: bool = False,
):
    batch_dim = jax.tree.leaves(xs)[0].shape[0]
    n_batches = batch_dim // batch_size
    return _batched_map(
        fun, xs, batch_dim, n_batches, batch_size, is_batch_axis_contracted
    )


def _batched_map(
    fun: Callable,
    xs: PyTree[Array],
    batch_dim: int,
    n_batches: int,
    batch_size: int,
    is_batch_axis_contracted: bool = False,
):
    """Like `jax.lax.map`, but map over leading axis of `xs` in
    chunks of size `batch_size`. Assumes `fun` can be evaluated in
    parallel over this leading axis.
    """
    # ... reshape into an iterative dimension and a batching dimension
    xs_per_batch = jax.tree.map(
        lambda x: x[: batch_dim - batch_dim % batch_size, ...].reshape(
            (n_batches, batch_size, *x.shape[1:])
        ),
        xs,
    )
    # .. compute the result and reshape back into one leading dimension
    result_per_batch = jax.lax.map(fun, xs_per_batch)
    if is_batch_axis_contracted:
        result = result_per_batch
    else:
        result = result_per_batch.reshape(
            (n_batches * batch_size, *result_per_batch.shape[2:])
        )
    # ... if the batch dimension is not divisible by the batch size, need
    # to take care of the remainder
    if batch_dim % batch_size != 0:
        remainder = fun(
            jax.tree.map(lambda x: x[batch_dim - batch_dim % batch_size :, ...], xs)
        )
        if is_batch_axis_contracted:
            remainder = remainder[None, ...]
        result = jnp.concatenate([result, remainder], axis=0)
    return result


def _batched_map_with_contraction(fun, xs, n_batches):
    # ... reshape into an iterative dimension and a batching dimension
    batch_dim = jax.tree.leaves(xs)[0].shape[0]
    batch_size = batch_dim // n_batches
    xs_per_batch = jax.tree.map(
        lambda x: x[: batch_dim - batch_dim % batch_size, ...].reshape(
            (n_batches, batch_size, *x.shape[1:])
        ),
        xs,
    )
    # .. compute the result and reshape back into one leading dimension
    result = jax.lax.map(fun, xs_per_batch)
    # ... if the batch dimension is not divisible by the batch size, need
    # to take care of the remainder
    if batch_dim % batch_size != 0:
        remainder = fun(
            jax.tree.map(lambda x: x[batch_dim - batch_dim % batch_size :, ...], xs)
        )[None, ...]
        result = jnp.concatenate([result, remainder], axis=0)
    return result



import cryojax.jax_util as jxu
import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest
from cryojax.io import read_array_from_mrc
from cryojax.ndimage import make_coordinate_grid
from jaxtyping import Array


#
# Test PyTree transforms
#
class Exp(eqx.Module):
    a: Array = eqx.field(converter=jnp.asarray)

    def __call__(self, x):
        return jnp.exp(-self.a * x)


def test_resolve_transform():
    pytree = Exp(a=1.0)
    pytree_with_transform = eqx.tree_at(
        lambda fn: fn.a,
        pytree,
        replace_fn=lambda a: jxu.CustomTransform(jnp.exp, jnp.log(a)),
    )
    assert eqx.tree_equal(pytree, jxu.resolve_transforms(pytree_with_transform))


def test_nested_resolve_transform():
    pytree = Exp(a=1.0)
    pytree_with_transform = eqx.tree_at(
        lambda fn: fn.a,
        pytree,
        replace_fn=lambda a: jxu.CustomTransform(lambda b: 2 * b, a / 2),
    )
    pytree_with_nested_transform = eqx.tree_at(
        lambda fn: fn.a.args[0],
        pytree_with_transform,
        replace_fn=lambda a_scaled: jxu.CustomTransform(jnp.exp, jnp.log(a_scaled)),
    )
    assert eqx.tree_equal(
        pytree,
        jxu.resolve_transforms(pytree_with_transform),
        jxu.resolve_transforms(pytree_with_nested_transform),
    )


def test_stop_gradient():
    @jax.value_and_grad
    def objective_fn(pytree):
        exp, x = jxu.resolve_transforms(pytree)
        return exp(x)

    x = jnp.asarray(np.random.random())
    exp = Exp(a=1.0)
    exp_with_stop_gradient = eqx.tree_at(
        lambda fn: fn.a, exp, replace_fn=jxu.StopGradientTransform
    )
    _, grads = objective_fn((exp_with_stop_gradient, x))
    grads = jxu.resolve_transforms(grads)
    assert grads[0].a == 0.0


#
# Test grid search
#
class ExampleModule(eqx.Module):
    a_1: Array
    a_2: Array
    a_3: Array
    placeholder: None

    def __init__(self, a_1, a_2, a_3):
        self.a_1 = a_1
        self.a_2 = a_2
        self.a_3 = a_3
        self.placeholder = None


def test_pytree_grid_manipulation():
    # ... make three arrays with the same leading dimension
    a_1, a_2, a_3 = tuple([jnp.arange(5) for _ in range(3)])
    # ... now two other arrays with different leading dimensions
    b, c = jnp.arange(7), jnp.arange(20)
    # Build a random tree grid
    is_leaf = lambda x: isinstance(x, ExampleModule)
    tree_grid = [ExampleModule(a_1, a_2, a_3), b, None, (c, (None,))]
    # Get grid point
    shape = jxu.tree_grid_shape(tree_grid, is_leaf=is_leaf)
    tree_grid_point = jxu.tree_grid_take(
        tree_grid, jxu.tree_grid_unravel_index(0, tree_grid, is_leaf=is_leaf)
    )
    tree_grid_points = jxu.tree_grid_take(
        tree_grid,
        jxu.tree_grid_unravel_index(jnp.asarray([0, 10]), tree_grid, is_leaf=is_leaf),
    )
    # Define ground truth
    true_shape = (a_1.size, b.size, c.size)
    true_tree_grid_point = [
        ExampleModule(a_1[0], a_2[0], a_3[0]),
        b[0],
        None,
        (c[0], (None,)),
    ]
    true_tree_grid_points = [
        ExampleModule(a_1[([0, 0],)], a_2[([0, 0],)], a_3[([0, 0],)]),
        b[([0, 0],)],
        None,
        (c[([0, 10],)], (None,)),
    ]
    assert shape == true_shape
    assert eqx.tree_equal(tree_grid_point, true_tree_grid_point)
    assert eqx.tree_equal(tree_grid_points, true_tree_grid_points)


@eqx.filter_jit
def cost_fn(grid_point, variance_plus_offset):
    variance, offset = variance_plus_offset
    mu_x, mu_y = offset
    x, y = grid_point
    return -jnp.exp(-((x - mu_x) ** 2 + (y - mu_y) ** 2) / (2 * variance)) / jnp.sqrt(
        2 * jnp.pi * variance
    )


@pytest.mark.parametrize(
    "batch_size,dim,offset,variance",
    [
        (None, 200, (-1.0, 2.0), 10.0),
        (1, 200, (-1.0, 2.0), 10.0),
        (10, 200, (-1.0, 2.0), 10.0),
        (33, 200, (99.0, 99.0), 10.0),
    ],
)
def test_run_grid_search(batch_size, dim, offset, variance):
    # Compute full landscape of simple analytic "cost function"
    coords = make_coordinate_grid((dim, dim))
    variance, offset = jnp.asarray(variance), jnp.asarray(offset)
    landscape = jax.vmap(jax.vmap(cost_fn, in_axes=[0, None]), in_axes=[0, None])(
        coords, (variance, offset)
    )
    # Find the true minimum value and its location
    true_min_eval = landscape.min()
    true_min_idx = jnp.squeeze(jnp.argwhere(landscape == true_min_eval))
    true_min_pos = tuple(coords[true_min_idx[0], true_min_idx[1]])
    # Generate a sparse representation of coordinate grid
    x, y = (
        jnp.fft.fftshift(jnp.fft.fftfreq(dim)) * dim,
        jnp.fft.fftshift(jnp.fft.fftfreq(dim)) * dim,
    )
    grid = (x, y)
    # Run the grid search
    method = jxu.MinimumSearchMethod(batch_size=batch_size)
    solution = jxu.run_grid_search(cost_fn, method, grid, (variance, offset))
    np.testing.assert_allclose(solution.state.current_minimum_eval, true_min_eval)
    np.testing.assert_allclose(solution.value, true_min_pos)


#
# Test `filter_bscan` / `filter_bmap`
#
@pytest.mark.parametrize(
    "batch_size,dim",
    [
        (1, 200),
        (10, 200),
        (33, 200),
        (200, 200),
    ],
)
def test_bscan_remainder(batch_size, dim):
    @jax.jit
    @jax.vmap
    def f(x):
        return x + 1

    xs = jnp.zeros(dim)
    np.testing.assert_allclose(jxu.filter_bmap(f, xs, batch_size=batch_size), f(xs))


#
# Linear operators
#
@pytest.fixture
def voxel_info(sample_mrc_path):
    return read_array_from_mrc(sample_mrc_path, loads_grid_spacing=True)


@pytest.fixture
def voxel_volume(voxel_info):
    return cxs.FourierVoxelGridVolume.from_real_voxel_grid(voxel_info[0], pad_scale=1.3)


@pytest.fixture
def voxel_size(voxel_info):
    return voxel_info[1]


@pytest.fixture
def image_config(voxel_volume, voxel_size):
    shape = voxel_volume.shape[0:2]
    return cxs.BasicImageConfig(
        shape=(int(0.9 * shape[0]), int(0.9 * shape[1])),
        pixel_size=voxel_size,
        voltage_in_kilovolts=300.0,
        pad_options=dict(shape=shape),
    )


@pytest.fixture
def image_model(voxel_volume, image_config):
    pose = cxs.EulerAnglePose()
    return cxs.make_image_model(voxel_volume, image_config, pose)


def test_simulate_equality(image_model):
    linear_operator, vector = jxu.make_linear_operator(
        fn=lambda x: x.simulate(),
        args=image_model,
        where_vector=lambda x: x.volume.fourier_voxel_grid,
    )
    image_cxs = image_model.simulate()
    image_lx = linear_operator.mv(vector)
    np.testing.assert_allclose(image_cxs, image_lx)


def test_linear_transpose(image_model):
    where_vector = lambda x: x.volume.fourier_voxel_grid
    linear_operator, _ = jxu.make_linear_operator(
        fn=lambda x: x.simulate(),
        args=image_model,
        where_vector=where_vector,
    )
    voxel_grid = where_vector(image_model)
    backprojection = where_vector(
        linear_operator.T.mv(jnp.zeros(image_model.image_config.shape))
    )
    assert voxel_grid.shape == backprojection.shape


def test_bad_linear_transpose(sample_pdb_path, image_config):
    image_model = cxs.make_image_model(
        cxs.load_tabulated_volume(sample_pdb_path, output_type=cxs.GaussianMixtureVolume),
        image_config,
        pose=cxs.EulerAnglePose(),
    )
    where_vector = lambda x: x.volume.positions
    linear_operator, _ = jxu.make_linear_operator(
        fn=lambda x: x.simulate(),
        args=image_model,
        where_vector=where_vector,
    )
    with pytest.raises(Exception):
        linear_operator.T



from typing import Literal

from jaxtyping import Array, Complex


def convert_fftn_to_rfftn(
    fftn_array: Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"],
    mode: Literal["zero", "one", "real"] | None = "zero",
) -> Complex[Array, "y_dim x_dim//2+1"] | Complex[Array, "z_dim y_dim x_dim//2+1"]:
    """Converts the output of a call to `jax.numpy.fft.fftn` to
    an `jax.numpy.fft.rfftn`.

    The FFT $F$ of a real-valued function obeys hermitian
    symmetry, i.e.

    $$F^*(k) = F(-k).$$

    Therefore, to convert an `fftn` output to that which would
    be returned by an `rfftn`, take the upper-half plane of
    an `fftn_array`. Also, optionally take care to make sure that
    self-conjugate components are purely real-valued.

    **Arguments:**

    - `fftn_array`:
        The output of a call to `jax.numpy.fft.fftn`.
    - `mode`:
        See the function`enforce_rfftn_self_conjugates`
        for documentation. If this is `None`, do not call this
        function.


    **Returns:**

    The `fftn_array`, as if it were the output of a call
    to `cryojax.image.rfftn` function.
    """
    shape = fftn_array.shape
    # Take upper half plane
    if fftn_array.ndim == 2:
        rfftn_array = fftn_array[:, : shape[-1] // 2 + 1]
    elif fftn_array.ndim == 3:
        rfftn_array = fftn_array[:, :, : shape[-1] // 2 + 1]
    else:
        raise NotImplementedError(
            "Only 2D and 3D arrays are supported "
            "in function `convert_fftn_to_rfftn`. "
            f"Passed an array with `ndim = {fftn_array.ndim}`."
        )
    if mode is not None:
        rfftn_array = enforce_rfftn_self_conjugates(
            rfftn_array,
            shape,  # type: ignore
            includes_dc=False,
            mode=mode,
        )
    return rfftn_array


def enforce_rfftn_self_conjugates(
    rfftn_array: (
        Complex[Array, "{shape[0]} {shape[1]}//2+1"]
        | Complex[Array, "{shape[0]} {shape[1]} {shape[2]}//2+1"]
    ),
    shape: tuple[int, int] | tuple[int, int, int],
    includes_dc: bool = False,
    mode: Literal["zero", "one", "real"] = "zero",
) -> (
    Complex[Array, "{shape[0]} {shape[1]}//2+1"]
    | Complex[Array, "{shape[0]} {shape[1]} {shape[2]}//2+1"]
):
    """For an array that is the output of a call to an "rfftn"
    function, enforce that self-conjugate components are real-valued.

    By default, do this by setting them to zero. This is important
    before applying translational phase shifts to an image in fourier space.

    **Arguments:**

    - `rfftn_array`:
        An array that is the output of a call to an
        "rfftn" function. This must have the zero-frequency
        component in the corner.
    - `shape`:
        The shape of the `rfftn_array` in real-space.
    - `includes_dc`:
        If `True`, enforce that `rfftn_array[0, 0]` is real.
        Otherwise, leave this component unmodified.
    - `mode`:
        A string controlling how the components are made
        real-valued. Supported modes are

        - "zero": sets components to zero
        - "one": sets components to one
        - "real": takes real part of components

        By default, `mode = "zero"`.

    **Return:**

    The modified `rfftn_array`, with self-conjugate components
    made real-valued.
    """
    if mode == "zero":
        replace_fn = lambda _: 0.0
    elif mode == "one":
        replace_fn = lambda _: 1.0
    elif mode == "real":
        replace_fn = lambda arr: arr.real
    else:
        raise NotImplementedError(
            f"`mode = {mode}` not supported for function "
            "`enforce_rfftn_self_conjugates`. "
            "The supported modes are 'zero', 'one', and 'real'."
        )
    if rfftn_array.ndim == 2:
        assert len(shape) == 2
        y_dim, x_dim = shape
        if includes_dc:
            rfftn_array = rfftn_array.at[0, 0].set(replace_fn(rfftn_array[0, 0]))
        if y_dim % 2 == 0:
            rfftn_array = rfftn_array.at[y_dim // 2, 0].set(
                replace_fn(rfftn_array[y_dim // 2, 0])
            )
        if x_dim % 2 == 0:
            rfftn_array = rfftn_array.at[0, x_dim // 2].set(
                replace_fn(rfftn_array[0, x_dim // 2])
            )
        if y_dim % 2 == 0 and x_dim % 2 == 0:
            rfftn_array = rfftn_array.at[y_dim // 2, x_dim // 2].set(
                replace_fn(rfftn_array[y_dim // 2, x_dim // 2])
            )
    elif rfftn_array.ndim == 3:
        assert len(shape) == 3
        z_dim, y_dim, x_dim = shape
        if includes_dc:
            rfftn_array = rfftn_array.at[0, 0, 0].set(replace_fn(rfftn_array[0, 0, 0]))
        if z_dim % 2 == 0:
            rfftn_array = rfftn_array.at[0, z_dim // 2, 0].set(
                replace_fn(rfftn_array[0, z_dim // 2, 0])
            )
        if y_dim % 2 == 0:
            rfftn_array = rfftn_array.at[0, y_dim // 2, 0].set(
                replace_fn(rfftn_array[0, y_dim // 2, 0])
            )
        if x_dim % 2 == 0:
            rfftn_array = rfftn_array.at[0, 0, x_dim // 2].set(
                replace_fn(rfftn_array[0, 0, x_dim // 2])
            )
        if y_dim % 2 == 0 and x_dim % 2 == 0:
            rfftn_array = rfftn_array.at[0, y_dim // 2, x_dim // 2].set(
                replace_fn(rfftn_array[0, y_dim // 2, x_dim // 2])
            )
        if z_dim % 2 == 0 and x_dim % 2 == 0:
            rfftn_array = rfftn_array.at[z_dim // 2, 0, x_dim // 2].set(
                replace_fn(rfftn_array[z_dim // 2, 0, x_dim // 2])
            )
        if z_dim % 2 == 0 and y_dim % 2 == 0:
            rfftn_array = rfftn_array.at[z_dim // 2, y_dim // 2, 0].set(
                replace_fn(rfftn_array[z_dim // 2, y_dim // 2, 0])
            )
        if z_dim % 2 == 0 and y_dim % 2 == 0 and x_dim % 2 == 0:
            rfftn_array = rfftn_array.at[z_dim // 2, y_dim // 2, x_dim // 2].set(
                replace_fn(rfftn_array[z_dim // 2, y_dim // 2, x_dim // 2])
            )
    else:
        raise NotImplementedError(
            "Only 2D and 3D arrays are supported "
            "in function `enforce_rfftn_self_conjugates`. "
            f"Passed an array with `ndim = {rfftn_array.ndim}`."
        )
    return rfftn_array



"""
Implementation of operators on images in real-space.
"""

from abc import abstractmethod
from collections.abc import Sequence
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Float, Inexact

from ..._internal import error_if_not_positive
from ...jax_util import FloatLike, NDArrayLike


class AbstractRealOperator(eqx.Module, strict=True):
    """
    The base class for all real operators.

    By convention, operators should be defined to
    have units of inverse area (up to a scale factor).

    To create a subclass,

        1) Include the necessary parameters in
           the class definition.
        2) Overrwrite the `__call__` method.
    """

    @abstractmethod
    def __call__(  # pyright: ignore
        self,
        coordinate_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        raise NotImplementedError

    def __add__(self, other) -> "AbstractRealOperator":
        if isinstance(other, AbstractRealOperator):
            return _SumRealOperator(self, other)
        return _SumRealOperator(self, RealConstant(other))

    def __radd__(self, other) -> "AbstractRealOperator":
        if isinstance(other, AbstractRealOperator):
            return _SumRealOperator(other, self)
        return _SumRealOperator(RealConstant(other), self)

    def __sub__(self, other) -> "AbstractRealOperator":
        if isinstance(other, AbstractRealOperator):
            return _DiffRealOperator(self, other)
        return _DiffRealOperator(self, RealConstant(other))

    def __rsub__(self, other) -> "AbstractRealOperator":
        if isinstance(other, AbstractRealOperator):
            return _DiffRealOperator(other, self)
        return _DiffRealOperator(RealConstant(other), self)

    def __mul__(self, other) -> "AbstractRealOperator":
        if isinstance(other, AbstractRealOperator):
            return _ProductRealOperator(self, other)
        return _ProductRealOperator(self, RealConstant(other))

    def __rmul__(self, other) -> "AbstractRealOperator":
        if isinstance(other, AbstractRealOperator):
            return _ProductRealOperator(other, self)
        return _ProductRealOperator(RealConstant(other), self)


class _SumRealOperator(AbstractRealOperator, strict=True):
    """A helper to represent the sum of two operators."""

    operator1: AbstractRealOperator
    operator2: AbstractRealOperator

    @override
    def __call__(
        self,
        coordinate_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.operator1(coordinate_grid) * self.operator2(coordinate_grid)

    def __repr__(self):
        return f"{repr(self.operator1)} + {repr(self.operator2)}"


class _DiffRealOperator(AbstractRealOperator, strict=True):
    """A helper to represent the difference of two operators."""

    operator1: AbstractRealOperator
    operator2: AbstractRealOperator

    @override
    def __call__(
        self,
        coordinate_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.operator1(coordinate_grid) * self.operator2(coordinate_grid)

    def __repr__(self):
        return f"{repr(self.operator1)} - {repr(self.operator2)}"


class _ProductRealOperator(AbstractRealOperator, strict=True):
    """A helper to represent the product of two operators."""

    operator1: AbstractRealOperator
    operator2: AbstractRealOperator

    @override
    def __call__(
        self,
        coordinate_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.operator1(coordinate_grid) * self.operator2(coordinate_grid)

    def __repr__(self):
        return f"{repr(self.operator1)} * {repr(self.operator2)}"


class RealGaussian(AbstractRealOperator, strict=True):
    """This operator is a normalized gaussian in real space

    $$g(r) = \\frac{\\kappa}{2\\pi \\beta} \\exp(- (r - r_0)^2 / (2 \\sigma))$$

    where $r^2 = x^2 + y^2$.
    """

    amplitude: Float[Array, ""]
    variance: Float[Array, ""]
    offset: Float[Array, " _"] | None

    def __init__(
        self,
        amplitude: FloatLike = 1.0,
        variance: FloatLike = 1.0,
        offset: Float[NDArrayLike, " _"] | Sequence[float] | None = None,
    ):
        """**Arguments:**

        - `amplitude`:
            The amplitude of the operator, equal to $\\kappa$
            in the above equation.
        - `variance`:
            The variance of the gaussian, equal to $\\sigma$
            in the above equation.
        - `offset`:
            An offset to the origin, equal to $r_0$
            in the above equation.
        """
        self.amplitude = jnp.asarray(amplitude, dtype=float)
        self.variance = jnp.asarray(variance, dtype=float)
        if offset is not None:
            self.offset = jnp.asarray(offset, dtype=float)
        else:
            self.offset = None

    @override
    def __call__(
        self, coordinate_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]:
        ndim = coordinate_grid.ndim - 1
        offset = jnp.zeros((ndim,), dtype=float) if self.offset is None else self.offset
        r_squared = jnp.sum((coordinate_grid - offset) ** 2, axis=-1)
        scaling = (
            self.amplitude
            / jnp.sqrt(2 * jnp.pi * error_if_not_positive(self.variance)) ** ndim
        ) * jnp.exp(-0.5 * r_squared / self.variance)
        return scaling


class RealConstant(AbstractRealOperator, strict=True):
    """An operator that is a constant."""

    value: Float[Array, "..."]

    def __init__(self, value: float | Float[NDArrayLike, "..."]):
        """**Arguments:**

        - `value`: The value of the constant
        """
        self.value = jnp.asarray(value)

    @override
    def __call__(
        self,
        coordinate_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, ""]:
        del coordinate_grid
        return self.value



"""
Routines for rescaling image pixel size.
"""

import jax
import jax.numpy as jnp
from jax.image import scale_and_translate
from jaxtyping import Array, Float


def rescale_pixel_size(
    image: Float[Array, "y_dim x_dim"],
    current_pixel_size: Float[Array, ""] | float,
    new_pixel_size: Float[Array, ""] | float,
    method: str = "bicubic",
    antialias: bool = True,
) -> Float[Array, "y_dim x_dim"]:
    """
    Measure an image at a given pixel size using interpolation.

    Parameters
    ----------
    image :
        The image to be magnified.
    current_pixel_size :
        The pixel size of the input image.
    new_pixel_size :
        The new pixel size after interpolation.
    method :
        Interpolation method. See ``jax.image.scale_and_translate``
        for documentation.
    antialias :
        Apply an anti-aliasing filter upon downsampling. See
        ``jax.image.scale_and_translate`` for documentation.


    Returns
    -------
    rescaled_image :
        An image with pixels whose size are rescaled by
        ``current_pixel_size / new_pixel_size``.
    """
    # Compute scale factor for pixel size rescaling
    scale_factor = current_pixel_size / new_pixel_size
    # Scaling in both dimensions is the same
    scaling = jnp.asarray([scale_factor, scale_factor])
    # Compute the translation in the jax.image convention that leaves
    # cryojax images untranslated
    N1, N2 = image.shape
    translation = (1 - scaling) * jnp.array([N1 // 2, N2 // 2], dtype=float)
    # Rescale pixel sizes
    rescaled_image = scale_and_translate(
        image,
        image.shape,
        (0, 1),
        scaling,
        translation,
        method,
        antialias=antialias,
        precision=jax.lax.Precision.HIGHEST,
    )

    return rescaled_image



import pathlib
from collections.abc import Callable
from typing import Any, Literal, overload

import equinox as eqx
import equinox.internal as eqxi
import jax.numpy as jnp
import mmdf
import pandas as pd
from jaxtyping import Bool

from ..atom_util import split_atoms_by_element
from ..constants import LobatoScatteringFactorParameters, PengScatteringFactorParameters
from ..io import mmdf_to_atoms
from ..jax_util import NDArrayLike
from ..ndimage import (
    AbstractImageTransform,
    compute_spline_coefficients,
    make_coordinate_grid,
    make_frequency_slice,
)
from ._detector import AbstractDetector
from ._image_config import AbstractImageConfig, DoseImageConfig
from ._image_model import (
    AbstractImageModel,
    ContrastImageModel,
    ElectronCountsImageModel,
    IntensityImageModel,
    LinearImageModel,
    ProjectionImageModel,
)
from ._pose import AbstractPose
from ._scattering_theory import WeakPhaseScatteringTheory
from ._transfer_theory import ContrastTransferTheory
from ._volume import (
    AbstractAtomVolume,
    AbstractVolumeIntegrator,
    AbstractVolumeParametrization,
    AbstractVolumeRenderFn,
    AutoVolumeProjection,
    FourierVoxelGridVolume,
    FourierVoxelSplineVolume,
    GaussianMixtureVolume,
    IndependentAtomVolume,
    RealVoxelGridVolume,
)


identity_fn = eqxi.doc_repr(lambda x, _: x, "identity_fn")


def _maybe_invert_rotation(
    pose: AbstractPose,
    volume: AbstractVolumeParametrization,
    rotation_convention: Literal["object", "frame"],
) -> AbstractPose:
    jaxpr_fn = eqx.filter_make_jaxpr(lambda vol: vol.to_representation())
    _, out_dynamic, out_static = jaxpr_fn(volume)
    out_struct = eqx.combine(out_dynamic, out_static)
    conventions = [rotation_convention, out_struct.rotation_convention]
    if conventions[0] != conventions[1]:
        pose = pose.to_inverse_rotation()

    return pose


@overload
def make_image_model(
    volume: AbstractVolumeParametrization,
    image_config: AbstractImageConfig,
    pose: AbstractPose,
    transfer_theory: None = None,
    volume_integrator: AbstractVolumeIntegrator = AutoVolumeProjection(),
    detector: AbstractDetector | None = None,
    *,
    image_transform: AbstractImageTransform | None = None,
    normalizes_signal: bool = False,
    signal_region: Bool[NDArrayLike, "_ _"] | None = None,
    signal_centering: Literal["bg", "mean"] = "mean",
    translate_mode: Literal["fft", "atom", "none"] = "fft",
    quantity_mode: None = None,
    rotation_convention: Literal["object", "frame"] = "object",
) -> ProjectionImageModel: ...


@overload
def make_image_model(  # pyright: ignore[reportOverlappingOverload]
    volume: AbstractVolumeParametrization,
    image_config: AbstractImageConfig,
    pose: AbstractPose,
    transfer_theory: ContrastTransferTheory,
    volume_integrator: AbstractVolumeIntegrator = AutoVolumeProjection(),
    detector: AbstractDetector | None = None,
    *,
    image_transform: AbstractImageTransform | None = None,
    normalizes_signal: bool = False,
    signal_region: Bool[NDArrayLike, "_ _"] | None = None,
    signal_centering: Literal["bg", "mean"] = "mean",
    translate_mode: Literal["fft", "atom", "none"] = "fft",
    quantity_mode: None = None,
    rotation_convention: Literal["object", "frame"] = "object",
) -> LinearImageModel: ...


@overload
def make_image_model(
    volume: AbstractVolumeParametrization,
    image_config: AbstractImageConfig,
    pose: AbstractPose,
    transfer_theory: ContrastTransferTheory,
    volume_integrator: AbstractVolumeIntegrator = AutoVolumeProjection(),
    detector: AbstractDetector | None = None,
    *,
    image_transform: AbstractImageTransform | None = None,
    normalizes_signal: bool = False,
    signal_region: Bool[NDArrayLike, "_ _"] | None = None,
    signal_centering: Literal["bg", "mean"] = "mean",
    translate_mode: Literal["fft", "atom", "none"] = "fft",
    quantity_mode: Literal["contrast"] = "contrast",
    rotation_convention: Literal["object", "frame"] = "object",
) -> ContrastImageModel: ...


@overload
def make_image_model(
    volume: AbstractVolumeParametrization,
    image_config: AbstractImageConfig,
    pose: AbstractPose,
    transfer_theory: ContrastTransferTheory,
    volume_integrator: AbstractVolumeIntegrator = AutoVolumeProjection(),
    detector: AbstractDetector | None = None,
    *,
    image_transform: AbstractImageTransform | None = None,
    normalizes_signal: bool = False,
    signal_region: Bool[NDArrayLike, "_ _"] | None = None,
    signal_centering: Literal["bg", "mean"] = "mean",
    translate_mode: Literal["fft", "atom", "none"] = "fft",
    quantity_mode: Literal["intensity"] = "intensity",
    rotation_convention: Literal["object", "frame"] = "object",
) -> IntensityImageModel: ...


@overload
def make_image_model(
    volume: AbstractVolumeParametrization,
    image_config: AbstractImageConfig,
    pose: AbstractPose,
    transfer_theory: ContrastTransferTheory,
    volume_integrator: AbstractVolumeIntegrator = AutoVolumeProjection(),
    detector: AbstractDetector | None = None,
    *,
    image_transform: AbstractImageTransform | None = None,
    normalizes_signal: bool = False,
    signal_region: Bool[NDArrayLike, "_ _"] | None = None,
    signal_centering: Literal["bg", "mean"] = "mean",
    translate_mode: Literal["fft", "atom", "none"] = "fft",
    quantity_mode: Literal["counts"] = "counts",
    rotation_convention: Literal["object", "frame"] = "object",
) -> ElectronCountsImageModel: ...


def make_image_model(
    volume: AbstractVolumeParametrization,
    image_config: AbstractImageConfig,
    pose: AbstractPose,
    transfer_theory: ContrastTransferTheory | None = None,
    volume_integrator: AbstractVolumeIntegrator = AutoVolumeProjection(),
    detector: AbstractDetector | None = None,
    *,
    image_transform: AbstractImageTransform | None = None,
    normalizes_signal: bool = False,
    signal_region: Bool[NDArrayLike, "_ _"] | None = None,
    signal_centering: Literal["bg", "mean"] = "mean",
    translate_mode: Literal["fft", "atom", "none"] = "fft",
    quantity_mode: Literal["contrast", "intensity", "counts"] | None = None,
    rotation_convention: Literal["object", "frame"] = "object",
) -> AbstractImageModel:
    """Construct an [`cryojax.simulator.AbstractImageModel`][] for
    most common use-cases.

    !!! example "Simulate an image"

        ```python
        import cryojax.simulator as cxs

        # Load modeling components
        volume, image_config, pose, transfer_theory = ...
        # Build image formation model
        image_model = cxs.make_image_model(volume, image_config, pose, transfer_theory)
        # Simulate!
        image = image_model.simulate()
        ```

    **Arguments:**

    - `volume`:
        The volume for imaging. To get started building volumes, see the
        [`cryojax.simulator.load_tabulated_volume`][] and
        [`cryojax.simulator.render_voxel_volume`][] utilities, or explore
        cryoJAX's [`cryojax.simulator.AbstractVolumeRepresentation`][] classes.
        Advanced users may be interested in implementing
        [`cryojax.simulator.AbstractVolumeParametrization`][] classes.
    - `image_config`:
        The configuration for the image and imaging instrument. Unless using
        a model that uses the electron dose as a parameter, choose the
        [`cryojax.simulator.BasicImageConfig`][]. Otherwise, choose the
        [`cryojax.simulator.DoseImageConfig`][].
    - `pose`:
        The pose in a particular parameterization convention. Common options
        are the [`cryojax.simulator.EulerAnglePose`][],
        [`cryojax.simulator.QuaternionPose`][], or
        [`cryojax.simulator.AxisAnglePose`][].
    - `transfer_theory`:
        The contrast transfer function and its theory for how it is applied
        to the image. See [`cryojax.simulator.ContrastTransferTheory`][].
    - `volume_integrator`:
        Optionally, pass the method for integrating the electrostatic potential onto
        the plane (e.g. projection via fourier slice extraction). If not provided,
        a default option is chosen using the
        [`cryojax.simulator.AutoVolumeProjection`][] class.
    - `detector`:
        If `quantity_mode = 'counts'` is chosen, then an
        [`cryojax.simulator.AbstractDetector`][] class must be chosen to
        simulate electron counts.
    - `image_transform`:
        A [`cryojax.ndimage.AbstractImageTransform`][] applied to the
        the output of `image_model.simulate()` as a postprocessing step.
    - `normalizes_signal`:
        Whether or not to normalize the output of `image_model.simulate()`.
        If `True`, see `signal_centering` for options.
    - `signal_region`:
        A boolean array that is 1 where there is signal,
        and 0 otherwise used to normalize the image.
        Must have shape equal to `image_config.shape`.
    - `signal_centering`:
        How to calculate the offset for normalization when
        `normalizes_signal = True`
        (and ignored if `normalizes_signal = False`).
        Options are
        - 'mean':
            Normalize the image to be mean 0
            within `signal_region`. This normalizes the image
            to be a z-score.
        - 'bg':
            Subtract mean value at the image edges.
            This makes the image fade to a background with values
            equal to zero. Requires that `image_config.padded_shape`
            is large enough so that the signal sufficiently decays.
    - `translate_mode`:
        How to apply in-plane translation to the volume. Options are
        - 'fft':
            Apply phase shifts in the Fourier domain.
        - 'atom':
            Apply translation to atom positions before
            projection. For this method, the `volume`
            argument must be or yield a [`cryojax.simulator.AbstractAtomVolume`][].
        - 'none':
            Do not apply the translation.
    - `quantity_mode`:
        The physical observable to simulate. If `None`, simulate without scaling
        to physical units using the [`cryojax.simulator.LinearImageModel`][].
        Options are
        - 'contrast':
            Uses the [`cryojax.simulator.ContrastImageModel`][]
            to simulate contrast.
        - 'intensity':
            Uses the [`cryojax.simulator.IntensityImageModel`][]
            to simulate intensity.
        - 'counts':
            Uses the [`cryojax.simulator.ElectronCountsImageModel`][]
            to simulate electron counts.
            If this is passed, a `detector` must also be passed.
    - `rotation_convention`:
        If `'object'`, the rotation given by `pose` is of the object.
        If `'frame'`, the rotation given by `pose` is of the frame. These
        are related by transpose.

    !!! info
        The `make_image_model` function enforces agreement between
        rotation conventions of different volumes via the
        `rotation_convention` argument. Lower level `cryojax` APIs
        will not enforce this agreement, such as if the user instantiates
        an [`cryojax.simulator.AbstractImageModel`][] directly.

        In these cases, agreement can be acheived with a manual transpose
        via `pose.to_inverse_rotation()`.

    **Returns:**

    An [`cryojax.simulator.AbstractImageModel`][]. This has type:

    - [`cryojax.simulator.ProjectionImageModel`][] if no `transfer_theory`
    is specified.
    - [`cryojax.simulator.LinearImageModel`][] if a `transfer_theory`
    is specified and `quantity_mode = None`.
    - [`cryojax.simulator.ContrastImageModel`][],
    [`cryojax.simulator.IntensityImageModel`][], or
    [`cryojax.simulator.ElectronCountsImageModel`][] depending on
    the value of `quantity_mode`.
    """
    # Invert pose if
    if rotation_convention not in ["object", "frame"]:
        raise ValueError(
            f"Found `rotation_convention = {rotation_convention}`, but valid "
            "values are 'object' and 'frame'."
        )
    pose = _maybe_invert_rotation(pose, volume, rotation_convention)
    options = dict(
        normalizes_signal=normalizes_signal,
        signal_centering=signal_centering,
        signal_region=signal_region,
        translate_mode=translate_mode,
        image_transform=image_transform,
    )
    if transfer_theory is None:
        # Image model for projections
        image_model = ProjectionImageModel(
            volume,
            pose,
            image_config,
            volume_integrator,
            **options,  # pyright: ignore[reportArgumentType]
        )
    else:
        # Simulate physical observables
        if quantity_mode is None:
            # Linear image model
            image_model = LinearImageModel(
                volume,
                pose,
                image_config,
                transfer_theory,
                volume_integrator,
                **options,  # pyright: ignore[reportArgumentType]
            )
        else:
            scattering_theory = WeakPhaseScatteringTheory(
                volume_integrator, transfer_theory
            )
            if quantity_mode == "counts":
                if not isinstance(image_config, DoseImageConfig):
                    raise ValueError(
                        "If using `quantity_mode = 'counts'` to simulate electron "
                        "counts, pass `image_config = DoseImageConfig(...)`. Got config "
                        f"{type(image_config).__name__}."
                    )
                if detector is None:
                    raise ValueError(
                        "If using `quantity_mode = 'counts'` to simulate electron "
                        "counts, an `AbstractDetector` must be passed."
                    )
                image_model = ElectronCountsImageModel(
                    volume,
                    pose,
                    image_config,
                    scattering_theory,
                    detector,
                    **options,  # pyright: ignore[reportArgumentType]
                )
            elif quantity_mode == "contrast":
                image_model = ContrastImageModel(
                    volume,
                    pose,
                    image_config,
                    scattering_theory,
                    **options,  # pyright: ignore[reportArgumentType]
                )
            elif quantity_mode == "intensity":
                image_model = IntensityImageModel(
                    volume,
                    pose,
                    image_config,
                    scattering_theory,
                    **options,  # pyright: ignore[reportArgumentType]
                )
            else:
                raise ValueError(
                    f"Found `quantity_mode = {quantity_mode}`, but valid "
                    "values are 'contrast', 'intensity', and 'counts'."
                )

    return image_model


@overload
def load_tabulated_volume(  # pyright: ignore[reportOverlappingOverload]
    path_or_mmdf: str | pathlib.Path | pd.DataFrame,
    *,
    output_type: type[IndependentAtomVolume] = IndependentAtomVolume,
    tabulation: Literal["peng", "lobato"] = "peng",
    include_b_factors: bool = True,
    b_factor_fn: Callable[[NDArrayLike, NDArrayLike], NDArrayLike] = identity_fn,
    selection_string: str = "all",
    pdb_options: dict[str, Any] = {},
) -> IndependentAtomVolume: ...


@overload
def load_tabulated_volume(
    path_or_mmdf: str | pathlib.Path | pd.DataFrame,
    *,
    output_type: type[GaussianMixtureVolume] = GaussianMixtureVolume,
    tabulation: Literal["peng"] = "peng",
    include_b_factors: bool = True,
    b_factor_fn: Callable[[NDArrayLike, NDArrayLike], NDArrayLike] = identity_fn,
    selection_string: str = "all",
    pdb_options: dict[str, Any] = {},
) -> GaussianMixtureVolume: ...


def load_tabulated_volume(
    path_or_mmdf: str | pathlib.Path | pd.DataFrame,
    *,
    output_type: type[
        IndependentAtomVolume | GaussianMixtureVolume
    ] = IndependentAtomVolume,
    tabulation: Literal["peng", "lobato"] = "peng",
    include_b_factors: bool = False,
    b_factor_fn: Callable[[NDArrayLike, NDArrayLike], NDArrayLike] = identity_fn,
    selection_string: str = "all",
    pdb_options: dict[str, Any] = {},
) -> IndependentAtomVolume | GaussianMixtureVolume:
    """Load an atomistic representation of a volume from
    tabulated electron scattering factors.

    !!! warning
        This function cannot be used with JIT compilation.
        Rather, its output should be passed to JIT-compiled
        functions. For example:

        ```python
        import cryojax.simulator as cxs
        import equinox as eqx

        path_to_pdb = ...
        volume = cxs.load_tabulated_volume(path_to_pdb)

        @eqx.filter_jit
        def simulate_fn(volume, ...):
            image_model = cxs.make_image_model(volume, ...)
            return image_model.simulate()

        image = simulate_fn(volume, ...)
        ```

    **Arguments:**

    - `path_or_mmdf`:
        The path to the PDB/PDBx file or a `pandas.DataFrame` loaded
        from [`mmdf.read`](https://github.com/teamtomo/mmdf).
    - `output_type`:
        Either a [`cryojax.simulator.GaussianMixtureVolume`][] or
        [`cryojax.simulator.IndependentAtomVolume`][] class.
    - `tabulation`:
        Specifies which electron scattering factor tabulation to use.
        Supported values are `tabulation = 'peng'` or `tabulation = 'lobato'`.
        See [`cryojax.constants.PengScatteringFactorParameters`][] and
        [`cryojax.constants.LobatoScatteringFactorParameters`][]
        for more information.
    - `include_b_factors`:
        If `True`, include PDB B-factors in the volume.
    - `b_factor_fn`:
        A function that modulates PDB B-factors before passing to the
        volume. Has signature
        `new_b_factor = b_factor_fn(b_factor, atomic_number)`.
        If `output_type = IndependentAtomVolume`, `b_factor` is
        the mean B-factor for a given atom type.
    - `selection_string`:
        A string for [`mdtraj` atom selection](https://mdtraj.org/1.9.4/examples/atom-selection.html#atom-selection).
        See [`cryojax.io.read_atoms_from_pdb`][] for documentation.
    - `pdb_options`:
        Additional keyword options passed to [`cryojax.io.read_atoms_from_pdb`][],
        not including `selection_string`.

    **Returns:**

    A [`cryojax.simulator.AbstractVoxelVolume`][] with exact type
    equal to `output_type`.
    """  # noqa: E501
    if isinstance(path_or_mmdf, (str, pathlib.Path)):
        atom_data = mmdf.read(pathlib.Path(path_or_mmdf))
    elif isinstance(path_or_mmdf, pd.DataFrame):
        atom_data = path_or_mmdf
    else:
        raise ValueError(
            "Argument `path_or_mmdf` to "
            "`load_tabulated_volume` was an unrecognized "
            "input type. Accepts a path to a PDB/PDBx file, "
            "or a pandas.DataFrame loaded from `mmdf.read`. "
            f"Instead, got type {path_or_mmdf.__class__.__name__}."
        )
    atom_positions, atomic_numbers, atom_properties = mmdf_to_atoms(
        atom_data,
        loads_properties=True,
        selection_string=selection_string,
        **pdb_options,
    )
    if output_type is GaussianMixtureVolume:
        if tabulation != "peng":
            raise ValueError(
                "Passed `output_type = GaussianMixtureVolume` to "
                "`load_tabulated_volume`, but found that "
                f"`tabulation = {tabulation}`, which "
                "is not a mixture of gaussians. Use "
                "`tabulation = 'peng'` instead."
            )
        peng_parameters = PengScatteringFactorParameters(atomic_numbers)
        b_factors = (
            jnp.asarray(
                b_factor_fn(atom_properties["b_factors"], atomic_numbers), dtype=float
            )
            if include_b_factors
            else None
        )
        atom_volume = GaussianMixtureVolume.from_tabulated_parameters(
            atom_positions, peng_parameters, extra_b_factors=b_factors
        )
    elif output_type is IndependentAtomVolume:
        (positions_by_id, b_factor_by_id), atom_ids = split_atoms_by_element(
            atomic_numbers, (atom_positions, atom_properties["b_factors"])
        )
        b_factor_by_id = tuple(
            jnp.asarray(b_factor_fn(jnp.mean(b), atom_ids)) for b in b_factor_by_id
        )
        if tabulation == "peng":
            parameters = PengScatteringFactorParameters(atom_ids)
        elif tabulation == "lobato":
            parameters = LobatoScatteringFactorParameters(atom_ids)
        else:
            raise ValueError(
                "Only `tabulation` equal to 'peng' or 'lobato' are supported in "
                f"`load_tabulated_volume`. Instead, got `tabulation = {tabulation}`."
            )
        atom_volume = IndependentAtomVolume.from_tabulated_parameters(
            positions_by_id, parameters, b_factor_by_element=b_factor_by_id
        )
    else:
        raise ValueError(
            "Only `output_type` equal to `GaussianMixtureVolume` "
            "or `IndependentAtomVolume` are supported."
        )

    return atom_volume


@overload
def render_voxel_volume(  # pyright: ignore[reportOverlappingOverload]
    atom_volume: AbstractAtomVolume,
    render_fn: AbstractVolumeRenderFn,
    *,
    output_type: type[FourierVoxelGridVolume] = FourierVoxelGridVolume,
) -> FourierVoxelGridVolume: ...


@overload
def render_voxel_volume(
    atom_volume: AbstractAtomVolume,
    render_fn: AbstractVolumeRenderFn,
    *,
    output_type: type[FourierVoxelSplineVolume] = FourierVoxelSplineVolume,
) -> FourierVoxelSplineVolume: ...


@overload
def render_voxel_volume(
    atom_volume: AbstractAtomVolume,
    render_fn: AbstractVolumeRenderFn,
    *,
    output_type: type[RealVoxelGridVolume] = RealVoxelGridVolume,
) -> RealVoxelGridVolume: ...


def render_voxel_volume(
    atom_volume: AbstractAtomVolume,
    render_fn: AbstractVolumeRenderFn,
    *,
    output_type: type[
        FourierVoxelGridVolume | FourierVoxelSplineVolume | RealVoxelGridVolume
    ] = FourierVoxelGridVolume,
) -> FourierVoxelGridVolume | FourierVoxelSplineVolume | RealVoxelGridVolume:
    """Render a voxel volume representation from an atomistic one.

    !!! example "Simulate an image with Fourier slice extraction"

        ```python
        import cryojax.simulator as cxs

        # Simulate an image with Fourier slice extraction
        voxel_volume = cxs.render_voxel_volume(
            atom_volume=cxs.load_tabulated_volume("example.pdb"),
            render_fn=cxs.AutoVolumeRenderFn(shape=(100, 100, 100), voxel_size=1.0),
            output_type=cxs.FourierVoxelGridVolume,
        )
        image_model = cxs.make_image_model(voxel_volume, ...)
        image = image_model.simulate()
        ```

    **Arguments:**

    - `atom_volume`:
        An atomistic volume representation, such as a
        [`cryojax.simulator.GaussianMixtureVolume`][] or a
        [`cryojax.simulator.IndependentAtomVolume`][].
    - `render_fn`:
        A [`cryojax.simulator.AbstractVolumeRenderFn`][] that
        accepts `atom_volume` as input. Choose
        [`cryojax.simulator.AutoVolumeRenderFn`][] to
        auto-select a method from existing cryoJAX
        implementations.
    - `output_type`:
        The [`cryojax.simulator.AbstractVoxelVolume`][]
        implementation to output.
        Either [`cryojax.simulator.FourierVoxelGridVolume`][] /
        [`cryojax.simulator.FourierVoxelSplineVolume`][] for
        fourier-space representations, or
        [`cryojax.simulator.RealVoxelGridVolume`][] for real-space.


    **Returns:**

    A [`cryojax.simulator.AbstractVoxelVolume`][] with exact type
    equal to `output_type`.
    """
    if len(set(render_fn.shape)) != 1:
        raise ValueError(
            "Function `render_voxel_volume` only supports "
            "volume rendering for cubic volumes, i.e. "
            "`render_fn.shape = (N, N, N)`. Got "
            f"`render_fn.shape = {render_fn.shape}`."
        )
    if output_type == FourierVoxelGridVolume or output_type == FourierVoxelSplineVolume:
        dim = render_fn.shape[0]
        frequency_slice = make_frequency_slice((dim, dim), outputs_rfftfreqs=False)
        fourier_voxel_grid = render_fn(
            atom_volume, outputs_real_space=False, outputs_rfft=False, fftshifted=True
        )
        if output_type == FourierVoxelGridVolume:
            return FourierVoxelGridVolume(fourier_voxel_grid, frequency_slice)
        else:
            spline_coefficients = compute_spline_coefficients(fourier_voxel_grid)
            return FourierVoxelSplineVolume(spline_coefficients, frequency_slice)
    elif output_type == RealVoxelGridVolume:
        coordinate_grid = make_coordinate_grid(render_fn.shape)
        real_voxel_grid = render_fn(atom_volume, outputs_real_space=True)
        return RealVoxelGridVolume(real_voxel_grid, coordinate_grid)
    else:
        raise ValueError(
            "Only `output_type` equal to `FourierVoxelGridVolume`, "
            "`FourierVoxelSplineVolume`, or `RealVoxelGridVolume` "
            "are supported."
            f"Got `output_type = {output_type}`."
        )



# Deprecation warnings
import warnings as _warnings
from typing import Any as _Any

from ._api_utils import (
    load_tabulated_volume as load_tabulated_volume,
    make_image_model as make_image_model,
    render_voxel_volume as render_voxel_volume,
)
from ._detector import (
    AbstractDetector as AbstractDetector,
    AbstractDQE as AbstractDQE,
    GaussianDetector as GaussianDetector,
    NullDQE as NullDQE,
    PoissonDetector as PoissonDetector,
)
from ._image_config import (
    AbstractImageConfig as AbstractImageConfig,
    BasicImageConfig as BasicImageConfig,
    DoseImageConfig as DoseImageConfig,
)
from ._image_model import (
    AbstractImageModel as AbstractImageModel,
    AbstractPhysicalImageModel as AbstractPhysicalImageModel,
    ContrastImageModel as ContrastImageModel,
    ElectronCountsImageModel as ElectronCountsImageModel,
    IntensityImageModel as IntensityImageModel,
    LinearImageModel as LinearImageModel,
    ProjectionImageModel as ProjectionImageModel,
)
from ._noise_model import (
    AbstractEmpiricalNoiseModel as AbstractEmpiricalNoiseModel,
    AbstractGaussianNoiseModel as AbstractGaussianNoiseModel,
    AbstractLikelihoodNoiseModel as AbstractLikelihoodNoiseModel,
    AbstractNoiseModel as AbstractNoiseModel,
    GaussianColoredNoiseModel as GaussianColoredNoiseModel,
    GaussianWhiteNoiseModel as GaussianWhiteNoiseModel,
)
from ._pose import (
    AbstractPose as AbstractPose,
    AxisAnglePose as AxisAnglePose,
    EulerAnglePose as EulerAnglePose,
    QuaternionPose as QuaternionPose,
)
from ._scattering_theory import (
    AbstractScatteringTheory as AbstractScatteringTheory,
    AbstractWaveScatteringTheory as AbstractWaveScatteringTheory,
    StrongPhaseScatteringTheory as StrongPhaseScatteringTheory,
    WeakPhaseScatteringTheory as WeakPhaseScatteringTheory,
)
from ._transfer_theory import (
    AbstractCTF as AbstractCTF,
    AbstractTransferTheory as AbstractTransferTheory,
    AstigmaticCTF as AstigmaticCTF,
    ContrastTransferTheory as ContrastTransferTheory,
    WaveTransferTheory as WaveTransferTheory,
)
from ._volume import (
    AbstractAtomVolume as AbstractAtomVolume,
    AbstractVolumeIntegrator as AbstractVolumeIntegrator,
    AbstractVolumeParametrization as AbstractVolumeParametrization,
    AbstractVolumeRenderFn as AbstractVolumeRenderFn,
    AbstractVolumeRepresentation as AbstractVolumeRepresentation,
    AbstractVoxelVolume as AbstractVoxelVolume,
    AutoVolumeProjection as AutoVolumeProjection,
    AutoVolumeRenderFn as AutoVolumeRenderFn,
    FFTAtomProjection as FFTAtomProjection,
    FFTAtomRenderFn as FFTAtomRenderFn,
    FourierSliceExtraction as FourierSliceExtraction,
    FourierVoxelGridVolume as FourierVoxelGridVolume,
    FourierVoxelSplineVolume as FourierVoxelSplineVolume,
    GaussianMixtureProjection as GaussianMixtureProjection,
    GaussianMixtureRenderFn as GaussianMixtureRenderFn,
    GaussianMixtureVolume as GaussianMixtureVolume,
    IndependentAtomVolume as IndependentAtomVolume,
    LobatoScatteringFactor as LobatoScatteringFactor,
    PengScatteringFactor as PengScatteringFactor,
    RealVoxelGridVolume as RealVoxelGridVolume,
    RealVoxelProjection as RealVoxelProjection,
)


def __getattr__(name: str) -> _Any:
    # Future deprecations
    if name == "AberratedAstigmaticCTF":
        _warnings.warn(
            "'AberratedAstigmaticCTF' is deprecated and will be removed in "
            "cryoJAX 0.6.0. Use 'AstigmaticCTF' instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        return AstigmaticCTF
    if name == "CTF":
        _warnings.warn(
            "Alias 'CTF' is deprecated and will be removed in "
            "cryoJAX 0.6.0. Use 'AstigmaticCTF' instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        return AstigmaticCTF
    if name == "NufftProjection":
        _warnings.warn(
            "'NufftProjection' is deprecated and will be removed in "
            "cryoJAX 0.6.0. Use 'RealVoxelProjection' instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        return RealVoxelProjection
    if name == "PengScatteringFactorParameters":
        _warnings.warn(
            "'PengScatteringFactorParameters' has been moved to `cryojax.constants` "
            "will be removed from `cryojax.simulator` in "
            "cryoJAX 0.6.0.",
            category=FutureWarning,
            stacklevel=2,
        )
        from ..constants import PengScatteringFactorParameters

        return PengScatteringFactorParameters
    if name == "PengAtomicVolume":
        _warnings.warn(
            "'PengAtomicVolume' is deprecated and will be removed in "
            "cryoJAX 0.6.0. To achieve identical functionality, use "
            "`GaussianMixtureVolume.from_tabulated_parameters`. "
            "This is a breaking change if you are "
            "directly using `PengAtomicVolume.__init__`.",
            category=FutureWarning,
            stacklevel=2,
        )
        return GaussianMixtureVolume
    if name == "UncorrelatedGaussianNoiseModel":
        _warnings.warn(
            "'UncorrelatedGaussianNoiseModel' is deprecated and "
            "will be removed in cryoJAX 0.6.0. Instead, use "
            "'GaussianWhiteNoiseModel'.",
            category=FutureWarning,
            stacklevel=2,
        )
        return GaussianWhiteNoiseModel
    if name == "CorrelatedGaussianNoiseModel":
        _warnings.warn(
            "'CorrelatedGaussianNoiseModel' is deprecated and "
            "will be removed in cryoJAX 0.6.0. Instead, use "
            "'GaussianColoredNoiseModel'.",
            category=FutureWarning,
            stacklevel=2,
        )
        return GaussianColoredNoiseModel
    # Deprecated in previous versions
    if name == "DiscreteStructuralEnsemble":
        raise ImportError(
            "'DiscreteStructuralEnsemble' was deprecated in cryoJAX 0.5.0. "
            "To achieve similar functionality, see the examples section "
            "of the documentation: "
            "https://michael-0brien.github.io/cryojax/examples/simulate-relion-dataset/.",
        )

    raise AttributeError(f"cannot import name '{name}' from 'cryojax.simulator'")



from typing import cast

import jax
import mdtraj
import mmdf
import numpy as np
import pytest
from cryojax.atom_util import center_atom_positions, split_atoms_by_element
from cryojax.io import mmdf_to_atoms, mmdf_to_topology, read_atoms_from_pdb


@pytest.mark.parametrize("mass_weighted, atom_axis", ([True, 0], [True, 1], [False, 0]))
def test_center_positions(sample_pdb_path, mass_weighted, atom_axis):
    # No batch dimension
    df = mmdf.read(sample_pdb_path)
    topology = mmdf_to_topology(df)
    atom_positions, _, atom_properties = mmdf_to_atoms(
        df, topology=topology, loads_properties=True
    )
    if atom_axis > 0:
        atom_positions, atom_properties = jax.tree.map(
            lambda x: np.expand_dims(x, axis=tuple(range(atom_axis))),
            (atom_positions, atom_properties),
        )
    # MDTraj centering
    traj = mdtraj.Trajectory(atom_positions / 10, topology)
    traj.center_coordinates(mass_weighted=mass_weighted)
    # cryoJAX centering
    cx_xyz = center_atom_positions(
        atom_positions,
        atom_masses=(atom_properties["masses"] if mass_weighted else None),
        atom_axis=atom_axis,
    )
    traj_xyz = np.expand_dims(cast(np.ndarray, traj.xyz)[0], axis=tuple(range(atom_axis)))

    np.testing.assert_allclose(cx_xyz, traj_xyz * 10, atol=1e-5)


@pytest.mark.parametrize("atom_axis", [1, 2])
def test_atom_splitting(sample_pdb_path, atom_axis):
    # No batch dimension
    atom_positions, atom_elements = read_atoms_from_pdb(
        sample_pdb_path, selection_string="element C || element O || element N"
    )
    positions_by_element, elements = split_atoms_by_element(
        atom_elements, atom_positions, atom_axis=0
    )
    assert isinstance(positions_by_element, tuple)
    assert len(positions_by_element) == 3
    assert len(elements) == 3

    batched_positions_by_element, _ = split_atoms_by_element(
        atom_elements,
        np.expand_dims(atom_positions, axis=tuple(range(atom_axis))),
        atom_axis=atom_axis,
    )
    assert all(
        pos.shape == bpos.shape[atom_axis:]
        for pos, bpos in zip(positions_by_element, batched_positions_by_element)
    )



import abc
import pathlib
from typing import Generic, Literal, TypeVar

import numpy as np
from jaxtyping import Float, Int, PyTree

from ...jax_util import NDArrayLike
from .._dataset import AbstractDataset


T1 = TypeVar("T1")
T2 = TypeVar("T2")


class AbstractParticleParameterFile(AbstractDataset[T1], Generic[T1, T2]):
    @abc.abstractmethod
    def __setitem__(self, index, value: T2):
        raise NotImplementedError

    @abc.abstractmethod
    def append(self, value: T2):
        raise NotImplementedError

    @abc.abstractmethod
    def save(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def path_to_output(self) -> pathlib.Path:
        raise NotImplementedError

    @path_to_output.setter
    @abc.abstractmethod
    def path_to_output(self, value: str | pathlib.Path):
        raise NotImplementedError

    @property
    def mode(self) -> Literal["r", "w"]:
        raise NotImplementedError


class AbstractParticleDataset(AbstractDataset[T1], Generic[T1, T2]):
    @property
    @abc.abstractmethod
    def parameter_file(self) -> AbstractParticleParameterFile:
        raise NotImplementedError

    @abc.abstractmethod
    def __setitem__(self, index, value: T2):
        raise NotImplementedError

    @abc.abstractmethod
    def append(self, value: T2):
        raise NotImplementedError

    @abc.abstractmethod
    def write_images(
        self,
        index_array: Int[np.ndarray, " _"],
        images: Float[NDArrayLike, "... _ _"],
        parameters: PyTree | None = None,
    ):
        raise NotImplementedError

    @property
    def mode(self) -> Literal["r", "w"]:
        raise NotImplementedError



from .transfer_function import (
    AbstractCTF as AbstractCTF,
    AstigmaticCTF as AstigmaticCTF,
)
from .transfer_theory import (
    AbstractTransferTheory as AbstractTransferTheory,
    ContrastTransferTheory as ContrastTransferTheory,
    WaveTransferTheory as WaveTransferTheory,
)



import cryojax.ndimage as im
import jax
import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest
from cryojax.ndimage import make_frequency_grid, make_radial_frequency_grid


#
# Downsampling
#
@pytest.mark.parametrize(
    "shape, downsample_factor",
    (((10, 10), 2), ((11, 11), 2)),
)
def test_downsample_preserves_sum(shape, downsample_factor):
    upsampled_shape = tuple(downsample_factor * s for s in shape)
    rng_key = jr.key(seed=1234)
    upsampled_image = 2.0 + 1.0 * jr.normal(rng_key, upsampled_shape)
    image = im.fourier_crop_downsample(upsampled_image, downsample_factor)
    np.testing.assert_allclose(image.sum(), upsampled_image.sum())


#
# FFT
#
@pytest.mark.parametrize("shape", [(10, 10), (10, 10, 10), (11, 11), (11, 11, 11)])
def test_fft_agrees_with_jax_numpy(shape):
    random = jnp.asarray(np.random.randn(*shape))
    # fftn
    np.testing.assert_allclose(random, im.ifftn(im.fftn(random)).real)
    np.testing.assert_allclose(
        im.ifftn(im.fftn(random)).real, jnp.fft.ifftn(jnp.fft.fftn(random)).real
    )
    # rfftn
    np.testing.assert_allclose(random, im.irfftn(im.rfftn(random), s=shape))
    np.testing.assert_allclose(
        im.irfftn(im.rfftn(random), s=shape),
        jnp.fft.irfftn(jnp.fft.rfftn(random), s=shape),
    )


#
# Cropping and padding
#
@pytest.mark.parametrize(
    "shape, cropped_shape, center, safe_crop",
    (
        ((10, 10), (5, 5), (5, 5), True),
        ((10, 10, 10), (5, 5, 5), (5, 5, 5), True),
        ((20, 20), (5, 5), (5, 5), True),
        ((20, 20), (5, 5), (5, 5), False),
        ((20, 20, 20), (5, 5, 5), (5, 5, 5), True),
        ((20, 20, 20), (5, 5, 5), (5, 5, 5), False),
        ((21, 21), (5, 5), (5, 5), True),
        ((21, 21), (5, 5), (5, 5), False),
        ((20, 20), (6, 6), (6, 6), True),
        ((21, 21), (6, 6), (6, 6), False),
    ),
)
def test_crop_with_center(shape, cropped_shape, center, safe_crop):
    coordinate_grid = im.make_radial_coordinate_grid(shape)
    if not safe_crop:
        center = tuple(jnp.asarray(c) for c in center)
    cropped_grid = im.crop_to_shape(coordinate_grid, cropped_shape, center=center)  # type: ignore
    assert cropped_grid.shape == cropped_shape


@pytest.mark.parametrize(
    "shape, cropped_shape",
    (
        ((10, 10), (5, 5)),
        ((10, 10), (6, 6)),
        ((11, 11), (5, 5)),
        ((11, 11), (6, 6)),
        ((11, 10), (5, 6)),
        ((10, 11), (6, 5)),
        ((11, 10), (6, 5)),
        ((10, 11), (5, 6)),
        ((10, 10, 10), (5, 5, 5)),
        ((10, 10, 10), (6, 6, 6)),
        ((11, 11, 11), (5, 5, 5)),
        ((11, 11, 11), (6, 6, 6)),
    ),
)
def test_crop(shape, cropped_shape):
    larger_frequency_grid = jnp.linalg.norm(
        jnp.asarray(shape, dtype=float)
        * jnp.fft.fftshift(make_frequency_grid(shape, outputs_rfftfreqs=False)),
        axis=-1,
    )
    smaller_frequency_grid = jnp.linalg.norm(
        jnp.asarray(cropped_shape, dtype=float)
        * jnp.fft.fftshift(make_frequency_grid(cropped_shape, outputs_rfftfreqs=False)),
        axis=-1,
    )
    cropped_frequency_grid = im.crop_to_shape(larger_frequency_grid, cropped_shape)
    dc_freq = tuple(jnp.asarray(s // 2, dtype=int) for s in cropped_shape)
    np.testing.assert_allclose(
        smaller_frequency_grid[dc_freq], cropped_frequency_grid[dc_freq]
    )
    np.testing.assert_allclose(smaller_frequency_grid, cropped_frequency_grid)


def test_crop_symmetric_signal():
    signal = np.zeros((20, 20))
    signal[0:7, 0:7] = 1.0
    signal[-7:, 0:7] = 1.0
    signal[0:7, -7:] = 1.0
    signal[-7:, -7:] = 1.0
    signal_crop = im.crop_to_shape(jnp.asarray(signal), (14, 14))
    np.testing.assert_allclose(np.sum(signal_crop[0:7, 0:7]), np.sum(signal_crop[7:, 7:]))


def test_pad_symmetric_signal():
    signal = np.zeros((20, 20))
    signal[0:7, 0:7] = 1.0
    signal[-7:, 0:7] = 1.0
    signal[0:7, -7:] = 1.0
    signal[-7:, -7:] = 1.0
    signal_pad = im.pad_to_shape(jnp.asarray(signal), (32, 32))
    np.testing.assert_allclose(
        np.sum(signal_pad[0:16, 0:16]), np.sum(signal_pad[16:, 16:])
    )


@pytest.mark.parametrize(
    "padded_shape, shape",
    (
        ((10, 10), (5, 5)),
        ((10, 10), (6, 6)),
        ((11, 11), (5, 5)),
        ((11, 11), (6, 6)),
        ((11, 10), (5, 6)),
        ((10, 11), (6, 5)),
        ((11, 10), (6, 5)),
        ((10, 11), (5, 6)),
        ((10, 10, 10), (5, 5, 5)),
        ((10, 10, 10), (6, 6, 6)),
        ((11, 11, 11), (5, 5, 5)),
        ((11, 11, 11), (6, 6, 6)),
    ),
)
def test_pad(padded_shape, shape):
    smaller_frequency_grid = jnp.linalg.norm(
        jnp.asarray(shape, dtype=float)
        * jnp.fft.fftshift(make_frequency_grid(shape, outputs_rfftfreqs=False)),
        axis=-1,
    )
    larger_frequency_grid = jnp.linalg.norm(
        jnp.asarray(padded_shape, dtype=float)
        * jnp.fft.fftshift(make_frequency_grid(padded_shape, outputs_rfftfreqs=False)),
        axis=-1,
    )
    padded_frequency_grid = im.pad_to_shape(smaller_frequency_grid, padded_shape)
    dc_freq = tuple(jnp.asarray(s // 2, dtype=int) for s in padded_shape)
    np.testing.assert_allclose(
        larger_frequency_grid[dc_freq], padded_frequency_grid[dc_freq]
    )
    np.testing.assert_allclose(
        im.crop_to_shape(larger_frequency_grid, shape),
        im.crop_to_shape(padded_frequency_grid, shape),
    )


# Normalization
def test_bg_subtract():
    image = jnp.ones((10, 10))
    image = np.asarray(im.background_subtract_image(image))
    assert np.all(image == 0.0)


# Fourier statistics
@pytest.mark.parametrize(
    "shape",
    [
        (10, 10),
        (10, 10, 10),
    ],
)
def test_powerspectrum_jit(shape):
    pixel_size = 1.2
    fourier_image = im.rfftn(jr.normal(jr.key(1234), shape))
    radial_frequency_grid = make_radial_frequency_grid(shape, pixel_size)

    @jax.jit
    def compute_powerspectrum_jit(image, radial_freqs, ps):
        return im.compute_binned_powerspectrum(
            image, radial_freqs, ps, minimum_frequency=0.0, maximum_frequency=0.5
        )

    try:
        _ = compute_powerspectrum_jit(fourier_image, radial_frequency_grid, pixel_size)
    except Exception as err:
        raise Exception(
            "Could not successfully run JIT compiled function "
            "`cryojax.image.compute_binned_powerspectrum`. "
            f"Error traceback was:\n{err}"
        )


@pytest.mark.parametrize(
    "shape",
    [
        (10, 10),
        (10, 10, 10),
    ],
)
def test_frc_fsc_jit(shape):
    if len(shape) == 2:
        correlation_fn = im.compute_fourier_ring_correlation
    else:
        correlation_fn = im.compute_fourier_shell_correlation
    pixel_size = 1.1
    fourier_image_1 = im.rfftn(jr.normal(jr.key(1234), shape))
    fourier_image_2 = im.rfftn(jr.normal(jr.key(2345), shape))
    radial_frequency_grid = make_radial_frequency_grid(shape, pixel_size)
    threshold = 0.5

    @jax.jit
    def compute_frc_fsc_jit(im1, im2, radial_freqs, ps, thresh):
        return correlation_fn(
            im1,
            im2,
            radial_freqs,
            ps,
            thresh,
            minimum_frequency=0.0,
            maximum_frequency=0.5,
        )

    try:
        _ = compute_frc_fsc_jit(
            fourier_image_1,
            fourier_image_2,
            radial_frequency_grid,
            pixel_size,
            threshold,
        )
    except Exception as err:
        raise Exception(
            "Could not successfully run JIT compiled function "
            f"`cryojax.image.{correlation_fn.__name__}`. "
            f"Error traceback was:\n{err}"
        )


# #
# # Pixel size rescaling
# #
# @pytest.mark.parametrize("shape", [(20, 20), (21, 21)])
# def test_rescale_pixel_size(shape):
#
#     image_1 = jr.normal(jr.key(0), shape)
#     pixel_size = 2.0
#     rescaled_pixel_size = 1.0
#     image_2 = im.rescale_pixel_size(
#         im.rescale_pixel_size(
#             image_1, pixel_size, rescaled_pixel_size, method="lanczos5"
#         ),
#         rescaled_pixel_size,
#         pixel_size,
#         method="lanczos5",
#     )
#     crop_1 = im.crop_to_shape(image_1, (shape[0] // 2, shape[1] // 2))
#     crop_2 = im.crop_to_shape(image_2, (shape[0] // 2, shape[1] // 2))
#     crop_2 = im.rescale_image(crop_2, crop_1.std(), crop_1.mean())
#     from matplotlib import pyplot as plt

#     # fig, axes = plt.subplots(ncols=2)
#     # vmin, vmax = min(crop_1.min(), crop_2.min()), max(crop_1.max(), crop_2.max())
#     # axes[0].imshow(crop_1, vmin=vmin, vmax=vmax)
#     # axes[1].imshow(crop_2, vmin=vmin, vmax=vmax)
#     # plt.show()
#     np.testing.assert_allclose(crop_1, crop_2)


def test_operators_instantiate():
    frequency_grid_2d = im.make_frequency_grid((10, 10))
    frequency_grid_3d = im.make_frequency_grid((10, 10, 10))
    coordinate_grid_2d = im.make_coordinate_grid((10, 10))
    coordinate_grid_3d = im.make_coordinate_grid((10, 10, 10))
    for cls in _real_operators_2d:
        _ = cls(coordinate_grid_2d)
    for cls in _real_operators_3d:
        _ = cls(coordinate_grid_3d)
    for cls in _fourier_operators_2d:
        _ = cls(frequency_grid_2d)
    for cls in _fourier_operators_3d:
        _ = cls(frequency_grid_3d)


_fourier_operators_2d = [
    im.FourierGaussian(),
    im.PeakedFourierGaussian(),
    im.FourierConstant(1.0),
    im.FourierSinc(),
    im.FourierExp2D(),
    im.CustomFourierOperator(lambda _, a, b: a + b, 1.0, b=1.0),
    im.FourierDC(),
    im.FourierConstant(1.0) + im.FourierConstant(1.0),
    im.FourierConstant(1.0) - im.FourierConstant(1.0),
    im.FourierConstant(1.0) * im.FourierConstant(1.0),
]


_fourier_operators_3d = [
    im.FourierGaussian(),
    im.PeakedFourierGaussian(),
    im.FourierConstant(1.0),
    im.FourierSinc(),
    im.CustomFourierOperator(lambda _, a, b: a + b, 1.0, b=1.0),
    im.FourierDC(),
    im.FourierConstant(1.0) + im.FourierConstant(1.0),
    im.FourierConstant(1.0) - im.FourierConstant(1.0),
    im.FourierConstant(1.0) * im.FourierConstant(1.0),
]

_real_operators_2d = [
    im.RealGaussian(),
    im.RealGaussian(offset=(1.0, -1.0)),
    im.RealConstant(1.0),
    im.RealConstant(1.0) + im.RealConstant(1.0),
    im.RealConstant(1.0) - im.RealConstant(1.0),
    im.RealConstant(1.0) * im.RealConstant(1.0),
]

_real_operators_3d = [
    im.RealGaussian(),
    im.RealGaussian(offset=(1.0, -1.0, 0.0)),
    im.RealConstant(1.0),
    im.RealConstant(1.0) + im.RealConstant(1.0),
    im.RealConstant(1.0) - im.RealConstant(1.0),
    im.RealConstant(1.0) * im.RealConstant(1.0),
]



import jax.numpy as jnp
from jaxtyping import Array, Float, Inexact


def cartesian_to_polar(
    coordinate_or_frequency_grid: Float[Array, "y_dim x_dim 2"], square: bool = False
) -> tuple[Inexact[Array, "y_dim x_dim"], Inexact[Array, "y_dim x_dim"]]:
    """Convert from cartesian to polar coordinates.

    **Arguments:**

    - `coordinate_or_frequency_grid`:
        The cartesian coordinate system in real or fourier space.
    - `square`:
        If `True`, return the square of the radial coordinate
        $|r|^2$. Otherwise, return $|r|$.

    **Returns:**

    A tuple `(r, theta)`, where `r` is the radial coordinate system and
    `theta` is the angular coordinate system. If `square=True`, return a
    tuple `(r_squared, theta)`.
    """
    theta = jnp.arctan2(
        coordinate_or_frequency_grid[..., 0], coordinate_or_frequency_grid[..., 1]
    )
    k_sqr = jnp.sum(jnp.square(coordinate_or_frequency_grid), axis=-1)
    if square:
        return k_sqr, theta
    else:
        kr = jnp.sqrt(k_sqr)
        return kr, theta



from typing import TypeAlias, TypeVar

from jaxtyping import Array, Int, PyTree, Shaped


SearchSolution = TypeVar("SearchSolution")
SearchState = TypeVar("SearchState")
PyTreeGrid: TypeAlias = PyTree[Shaped[Array, "_ ..."] | None, " Y"]
PyTreeGridPoint: TypeAlias = PyTree[Shaped[Array, "..."] | None, " Y"]
PyTreeGridIndex: TypeAlias = PyTree[Int[Array, "..."] | None, "... Y"]



from abc import abstractmethod
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, Inexact, PRNGKeyArray

from .._internal import error_if_not_fractional
from ..jax_util import FloatLike
from ..ndimage import fftn, ifftn, irfftn, rfftn
from ._image_config import AbstractImageConfig
from ._solvent_2d import AbstractRandomSolvent2D
from ._transfer_theory import ContrastTransferTheory, WaveTransferTheory
from ._volume import AbstractVolumeIntegrator, AbstractVolumeRepresentation


class AbstractScatteringTheory(eqx.Module, strict=True):
    """Base class for a scattering theory."""

    @abstractmethod
    def compute_contrast_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
        defocus_offset: FloatLike | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        raise NotImplementedError

    @abstractmethod
    def compute_intensity_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
        defocus_offset: FloatLike | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        raise NotImplementedError


class AbstractWaveScatteringTheory(AbstractScatteringTheory, strict=True):
    """Base class for a wave-based scattering theory."""

    transfer_theory: eqx.AbstractVar[WaveTransferTheory]

    @abstractmethod
    def compute_exit_wave(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]:
        raise NotImplementedError

    @override
    def compute_intensity_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
        defocus_offset: FloatLike | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        # ... compute the exit wave
        fourier_wavefunction = fftn(
            self.compute_exit_wave(volume_representation, image_config, rng_key)
        )
        # ... propagate to the detector plane
        fourier_wavefunction = self.transfer_theory.propagate_exit_wave(
            fourier_wavefunction,
            image_config,
            defocus_offset=defocus_offset,
        )
        wavefunction = ifftn(fourier_wavefunction)
        # ... get the squared wavefunction and return to fourier space
        intensity_spectrum = rfftn((wavefunction * jnp.conj(wavefunction)).real)

        return intensity_spectrum

    @override
    def compute_contrast_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
        defocus_offset: FloatLike | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        """Compute the contrast at the detector plane, given the squared wavefunction."""
        # ... compute the exit wave
        fourier_wavefunction = fftn(
            self.compute_exit_wave(volume_representation, image_config, rng_key)
        )
        # ... propagate to the detector plane
        fourier_wavefunction = self.transfer_theory.propagate_exit_wave(
            fourier_wavefunction,
            image_config,
            defocus_offset=defocus_offset,
        )
        wavefunction = ifftn(fourier_wavefunction)
        # ... get the squared wavefunction
        squared_wavefunction = (wavefunction * jnp.conj(wavefunction)).real
        # ... compute the contrast directly from the squared wavefunction
        # as C = -1 + psi^2 / 1 + psi^2
        contrast_spectrum = rfftn(
            (-1 + squared_wavefunction) / (1 + squared_wavefunction)
        )

        return contrast_spectrum


class WeakPhaseScatteringTheory(AbstractScatteringTheory, strict=True):
    """Base linear image formation theory."""

    volume_integrator: AbstractVolumeIntegrator
    transfer_theory: ContrastTransferTheory
    solvent: AbstractRandomSolvent2D | None = None

    def __init__(
        self,
        volume_integrator: AbstractVolumeIntegrator,
        transfer_theory: ContrastTransferTheory,
        solvent: AbstractRandomSolvent2D | None = None,
    ):
        """**Arguments:**

        - `volume_integrator`: The method for integrating the scattering potential.
        - `transfer_theory`: The contrast transfer theory.
        - `solvent`: The model for the solvent.
        """
        self.volume_integrator = volume_integrator
        self.transfer_theory = transfer_theory
        self.solvent = solvent

    def compute_object_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        # Compute the integrated potential
        fourier_in_plane_potential = self.volume_integrator.integrate(
            volume_representation, image_config, outputs_real_space=False
        )

        if rng_key is not None:
            # Get the potential of the specimen plus the ice
            if self.solvent is not None:
                fourier_in_plane_potential = self.solvent.compute_in_plane_potential(  # noqa: E501
                    rng_key,
                    fourier_in_plane_potential,
                    image_config,
                    input_is_rfft=not self.volume_integrator.outputs_ewald_sphere,
                )

        object_spectrum = image_config.interaction_constant * fourier_in_plane_potential

        return object_spectrum

    @override
    def compute_contrast_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
        defocus_offset: FloatLike | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        object_spectrum = self.compute_object_spectrum(
            volume_representation, image_config, rng_key
        )
        contrast_spectrum = self.transfer_theory.propagate_object(  # noqa: E501
            object_spectrum,
            image_config,
            input_is_ewald_sphere=self.volume_integrator.outputs_ewald_sphere,
            defocus_offset=defocus_offset,
        )

        return contrast_spectrum

    @override
    def compute_intensity_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
        defocus_offset: FloatLike | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        """Compute the squared wavefunction at the detector plane, given the
        contrast.
        """
        N1, N2 = image_config.padded_shape
        # ... compute the squared wavefunction directly from the image contrast
        # as |psi|^2 = 1 + 2C.
        contrast_spectrum = self.compute_contrast_spectrum(
            volume_representation, image_config, rng_key, defocus_offset=defocus_offset
        )
        intensity_spectrum = (2 * contrast_spectrum).at[0, 0].add(1.0 * N1 * N2)
        return intensity_spectrum


class StrongPhaseScatteringTheory(AbstractWaveScatteringTheory, strict=True):
    """Scattering theory for strong phase objects. This is analogous to
    a Moliere high-energy approximation in high-energy physics.

    This is the simplest model for multiple scattering events.

    !!! info
        Unlike in the weak-phase approximation, it is not possible to absorb a model
        for amplitude contrast (here via the amplitude contrast ratio) into the CTF.
        Instead, it is necessary to compute a complex scattering potential, where the
        imaginary part captures inelastic scattering.

        In particular, given a projected electrostatic potential $\\u(x, y)$, the
        complex potential $\\phi(x, y)$ for amplitude contrast ratio $\\alpha$ is

        $$\\phi(x, y) = \\sqrt{1 - \\alpha^2} \\ u(x, y) + i \\alpha \\ u(x, y).$$

    **References:**

    - See Chapter 69, Page 2012, from *Hawkes, Peter W., and Erwin Kasper.
      Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic
      Press, 2022.*
    - See Section 3.4, Page 61, from *Spence, John CH. High-resolution electron
      microscopy. OUP Oxford, 2013.*
    """

    volume_integrator: AbstractVolumeIntegrator
    transfer_theory: WaveTransferTheory
    solvent: AbstractRandomSolvent2D | None
    amplitude_contrast_ratio: Float[Array, ""]

    def __init__(
        self,
        volume_integrator: AbstractVolumeIntegrator,
        transfer_theory: WaveTransferTheory,
        solvent: AbstractRandomSolvent2D | None = None,
        amplitude_contrast_ratio: FloatLike = 0.1,
    ):
        """**Arguments:**

        - `volume_integrator`: The method for integrating the scattering potential.
        - `transfer_theory`: The wave transfer theory.
        - `solvent`: The model for the solvent.
        - `amplitude_contrast_ratio`: The amplitude contrast ratio.
        """
        self.volume_integrator = volume_integrator
        self.transfer_theory = transfer_theory
        self.solvent = solvent
        self.amplitude_contrast_ratio = jnp.asarray(amplitude_contrast_ratio, dtype=float)

    @override
    def compute_exit_wave(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]:
        # Compute the integrated potential in the exit plane
        fourier_in_plane_potential = self.volume_integrator.integrate(
            volume_representation, image_config, outputs_real_space=False
        )
        # The integrated potential may not be from an rfft; this depends on
        # if it is a projection approx
        is_projection_approx = not self.volume_integrator.outputs_ewald_sphere
        if rng_key is not None:
            # Get the potential of the specimen plus the ice
            if self.solvent is not None:
                fourier_in_plane_potential = self.solvent.compute_in_plane_potential(
                    rng_key,
                    fourier_in_plane_potential,
                    image_config,
                    input_is_rfft=is_projection_approx,
                )
        # Back to real-space; need to be careful if the object spectrum is not an
        # rfftn
        do_ifft = lambda ft: (
            irfftn(ft, s=image_config.padded_shape)
            if is_projection_approx
            else ifftn(ft, s=image_config.padded_shape)
        )
        integrated_potential = _compute_complex_potential(
            do_ifft(fourier_in_plane_potential),
            error_if_not_fractional(self.amplitude_contrast_ratio),
        )
        object = image_config.interaction_constant * integrated_potential
        # Compute wavefunction, with amplitude and phase contrast
        return jnp.exp(1.0j * object)


def _compute_complex_potential(
    in_plane_potential: Inexact[Array, "y_dim x_dim"],
    amplitude_contrast_ratio: Float[Array, ""] | float,
) -> Complex[Array, "y_dim x_dim"]:
    ac = amplitude_contrast_ratio
    if jnp.iscomplexobj(in_plane_potential):
        raise NotImplementedError(
            "You may have tried to use a `StrongPhaseScatteringTheory` "
            "together with `EwaldSphereExtraction` for simulating images. "
            "This is not implemented!"
        )
        # return jnp.sqrt(1.0 - ac**2) * integrated_potential.real + 1.0j * (
        #     integrated_potential.imag + ac * integrated_potential.real
        # )
    else:
        return (jnp.sqrt(1.0 - ac**2) + 1.0j * ac) * in_plane_potential



from .auto_select import (
    AutoVolumeProjection as AutoVolumeProjection,
    AutoVolumeRenderFn as AutoVolumeRenderFn,
)
from .base_volume import (
    AbstractAtomVolume as AbstractAtomVolume,
    AbstractVolumeIntegrator as AbstractVolumeIntegrator,
    AbstractVolumeParametrization as AbstractVolumeParametrization,
    AbstractVolumeRenderFn as AbstractVolumeRenderFn,
    AbstractVolumeRepresentation as AbstractVolumeRepresentation,
    AbstractVoxelVolume as AbstractVoxelVolume,
)
from .fourier_voxels import (
    EwaldSphereExtraction as EwaldSphereExtraction,
    FourierSliceExtraction as FourierSliceExtraction,
    FourierVoxelGridVolume as FourierVoxelGridVolume,
    FourierVoxelSplineVolume as FourierVoxelSplineVolume,
)
from .gaussian_volume import (
    GaussianMixtureProjection as GaussianMixtureProjection,
    GaussianMixtureRenderFn as GaussianMixtureRenderFn,
    GaussianMixtureVolume as GaussianMixtureVolume,
)
from .independent_atom_volume import (
    FFTAtomProjection as FFTAtomProjection,
    FFTAtomRenderFn as FFTAtomRenderFn,
    IndependentAtomVolume as IndependentAtomVolume,
    LobatoScatteringFactor as LobatoScatteringFactor,
    PengScatteringFactor as PengScatteringFactor,
)
from .real_voxels import (
    RealVoxelGridVolume as RealVoxelGridVolume,
    RealVoxelProjection as RealVoxelProjection,
)
